/*
    \brief  Deklarationen der Funktionen für die Arbeit mit den Setuplisten.

    \author Bernd Ottmann
    \date   21.02.2003
*/
#pragma once
#ifndef DATAFOX_SETUP_LISTS_H
#ifndef DOXYGEN_SHOULD_SKIP_THIS
#define DATAFOX_SETUP_LISTS_H
#endif

#include "df_defines.h"

/*!
    \brief Bibliotheksinterne Datenpuffer für Setuplisten löschen und initalisieren.

    Die in der Bibliothek bereitgestellten Datenpuffer für die zu übertragenenden Listendaten werden gelöscht und
    für eine Verwendung mittels DFCMakeListe() und DFCLoadListen() initalisiert.

    \ingroup setup_lists

    \export{_DFCClrListenBuffer@4, DFCClrListenBuffer}

    \param[in] connectionId Zu verwendende Verbindungs-Id.
*/
DF_API_EXTERN_DECORATE void DF_API_CALL DFCClrListenBuffer( int connectionId );

/*!
    \brief Kopieren eines aufbereiteten Datenarrays für eine Setupliste in den Biblioteksinternen Datenpuffer.

    Das zu übergebende Datenarray wird in den Bibliotheksinternen Datenpuffer kopiert und für eine spätere
    Übertragung mittels DFCLoadListen() zwischengespeichert.

    \ingroup setup_lists

    \export{_DFCMakeListe@24, DFCMakeListe}

    \param[in] connectionId Zu verwendende Verbindungs-Id.
    \param[in] number Listennnummer.
    \param[in] recordCount Anzahl der Datensätze im übergebenen Datenarray.
    \param[in] size Anzahl Byte des Datenarrays.\n Brechnet sich aus (recordCount * Größe pro Datensatz).
    \param[in] buffer Datenarray mit den Daten.
    \param[in] unused_setTo0 Sollte mit 0 übergeben werden. Reservierter Parameter.

    \retval 0 Fehler bei Ausführung.
    \retval 1 Ausführung war erfolgreich.
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCMakeListe( int connectionId, int number, int recordCount, int size, const unsigned char *buffer, int unused_setTo0 );

/*!
    \brief Übertragen der bibliotheksinternen Datenpuffer zum Gerät.

    Es werden die mit DFCMakeListe() importierten Listendaten in das Gerät geschrieben. Vorhandene Listendaten
    werden hierbei erst nach einer erfolgreichen Übernahme überschrieben.

    \ingroup setup_lists

    \export{_DFCLoadListen@12, DFCLoadListen}

    \attention Bitte beachten Sie die [Hinweise zur Lebensdauer](\ref dfc_lists_memory) der eingesetzten Flash-Bausteine.
    Diese gelten ebenfalls für Zutrittslisten.

    \param[in] connectionId Zu verwendende Verbindungs-Id.
    \param[in] deviceAddress Anzusprechende Geräteadresse.
    \param[out] errorNumber Fehlernummer bei fehlerhafter Ausführung.

    \retval 0 Fehler bei Ausführung.
    \retval 1 Ausführung war erfolgreich.
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCLoadListen( int connectionId, int deviceAddress, int *errorNumber );

/*!
    \brief Aufbau der Listentabellen aus dem Geräte lesen.

    Es werden die im Gerätesetup verwendeten Tabellenbeschreibungen für die Setuplisten in den Bibliotheksinternen
    Puffer gelesen. Hiernach kann mit den Funktionen DFCListBCnt(), DFCListBDatensatz() und DFCListBFeld()
    auf die ermittelten Daten zugegriffen werden.

    \ingroup setup_lists

    \export{_DFCLoadListenbeschreibung@12, DFCLoadListenbeschreibung}

    \param[in] connectionId Zu verwendende Verbindungs-Id.
    \param[in] deviceAddress Anzusprechende Geräteadresse.
    \param[out] errorNumber Fehlernummer bei fehlerhafter Ausführung.

    \retval 0 Fehler bei Ausführung.
    \retval 1 Ausführung war erfolgreich.
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCLoadListenbeschreibung( int connectionId, int deviceAddress, int *errorNumber );

/*!
    \brief Anzahl der gelesenen Listentabellen abrufen.

    Auf Grundlage der zuvor mit DFCLoadListenbeschreibung() gelesenen Listentabellen, wird
    die im Gerätesetup definierte Anzahl zurückgegeben.

    \ingroup setup_lists

    \export{_DFCListBCnt@4, DFCListBCnt}

    \param[in] connectionId Zu verwendende Verbindungs-Id.

    \return Anzahl der Listentabellen.
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCListBCnt( int connectionId );

/*!
    \brief Grunddaten einer Listentabelle abrufen.

    Auf Grundlage der zuvor mit DFCLoadListenbeschreibung() gelesenen Listentabellen, werden
    die Grunddaten zu einer Listentabelle zurückgegeben.

    \ingroup setup_lists

    \export{_DFCListBDatensatz@20, DFCListBDatensatz}

    \param[in] connectionId Zu verwendende Verbindungs-Id.
    \param[in] number Listennummer. Bereich 0 bis DFCListBCnt() - 1
    \param[out] name Zeichenpuffer für den Namen. Muss mindestens 17 Zeichen lang sein.
    \param[out] fieldCount Anzahl der Felder.
    \param[out] unused_setTo0 Variable welche eine 0 erhält. Reservierter Parameter.

    \retval 0 Fehler bei Ausführung.
    \retval 1 Ausführung war erfolgreich.
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCListBDatensatz( int connectionId, int number, char *name, int *fieldCount, int *unused_setTo0 );

/*!
    \brief Grunddaten eines Listenfeldes abrufen.

    Auf Grundlage der zuvor mit DFCLoadListenbeschreibung() gelesenen Listentabellen, werden
    die Grunddaten zu einem Listenfeld zurückgegeben.

    \ingroup setup_lists

    \export{_DFCListBFeld@24, DFCListBFeld}

    \param[in] connectionId Zu verwendende Verbindungs-Id.
    \param[in] listBNumber Tabellennummer im Bereich von (0 - (DFCListBCnt() - 1)).
    \param[in] number Feldnummer. Bereich 0 bis maximal 25.
    \param[out] name Zeichenpuffer für den Feldnamen. Muss mindestens 17 Zeichen lang sein.
    \param[out] type Datentyp.
    <table class="paramvalues">
        <tr><th>Wert</th><th>Bedeutung</th></tr>
        <tr><td>1</td><td>unsigned long 4 Bytes</td></tr>
        <tr><td>2</td><td>Date und Time 7 Bytes</td></tr>
        <tr><td>3</td><td>Wie Typ 4, es sind jedoch nur Ziffern erlaubt</td></tr>
        <tr><td>4</td><td>Alpha-Numerisch ASCII</td></tr>
        <tr><td>5</td><td>Datum 4 Bytes</td></tr>
        <tr><td>6</td><td>Zeit 3 Bytes</td></tr>
    </table>
    \param[out] size Länge. Die Längenangabe beinhaltet ein eventuell benötigtes Nullterminierungszeichen bei Feldern
    für Zeichenfolgen.

    \retval 0 Fehler bei Ausführung.
    \retval 1 Ausführung war erfolgreich.
*/
DF_API_EXTERN_DECORATE int DF_API_CALL DFCListBFeld( int connectionId, int listBNumber, int number, char *name, int *type, int *size );

#endif
