/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/

#include "channel/channel_bridge.h"
#include "DatafoxLibraryIV/include/library_version.h"
#include "DatafoxLibraryIV/include/datafox_library_iv.h"
#include "common/eeprom.h"
#include "common/crc.h"
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#if defined(_MSC_VER) || defined(__MINGW32__)
#include <io.h>
#else
#include <sys/uio.h>
#endif
#include "file/file_studio_setup_xml.h"

#if defined(_MSC_VER) && !defined(snprintf)
#define snprintf _snprintf
#endif

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

ChannelBridge::ChannelBridge( int connectionId )
    : CDFLog( connectionId, std::bind( &ChannelBridge::change_trace_file_path, this, std::placeholders::_1 ) )
{
    // Kanalnummer setzen
    setConnectionId( connectionId );
    
    // Protokollzustaende der angesprochenen Geraete
    memset( protocolVersion_, 0, sizeof(protocolVersion_) );

    // Bevorzugt AEIII+ pruefen
    preferAEIII_ = false;

    // Protocolversion für späteren Pufferzugriff halten.
    protocolVersionRecordDescription_ = 4;
    protocolVersionListDescription_ = 4;

    // Initialisierungsdatei lesen
    logIniRead(); // führt auch logCheck durch.

    // Ist die bevorzugte Version der Dockingstation == 1, dann wird diese Klasse aktiviert.
    channel4_.get_protocol().set_timeboy_station_enabled( m_docking_prefer_type == 1 );

    // Datensatzpuffer loeschen
    m_record_size = 0;
    m_record_error  = 0;
    memset( m_record_data, 0, sizeof(m_record_data) );

    // Wiederholungsversuche bei checkDevice
    devicePollRetry_ = 0;

    // Info zum binden des Kanals an ein bestimmtes Geraet.
    memset( &bindInfo_, 0, sizeof(bindInfo_) );
    bindInfo_.typeDevice = -1; // 0 wuerde PZE-MasterIV sein. -1 ist kein Geraet.

    channel4_.set_trace_mode( static_cast<Trace_base::Trace_mode>( m_trace_mode ) );
}

ChannelBridge::~ChannelBridge()
{

}

bool ChannelBridge::tryEnterCriticalSection()
{
    return mutexAPI_.try_lock();
}

void ChannelBridge::enterCriticalSection()
{
    mutexAPI_.lock();
}

void ChannelBridge::leaveCriticalSection()
{
    mutexAPI_.unlock();
}

void ChannelBridge::change_trace_file_path( const std::string& file_path )
{
    channel4_.set_trace_file_path( file_path );
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
int ChannelBridge::openAEIII( int iType, const char *szComString, int iComValue, int iComTimeOut )
{
    // Bevorzugt AEIII+ pruefen
    preferAEIII_ = true;

    return open( Protocol_base::dv_device_address, iType, szComString, iComValue, iComTimeOut );
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
int ChannelBridge::open( int deviceAddress, int iType, const std::string &szComString, int iComValue, int iComTimeOut )
{
    int res = IV_ERROR_FAIL;

    // Kanal darf nicht gebunden sein.
    if ( isBind() ) 
        return IV_ERROR_FAIL;

    // Kanal 4 ist Massgebend
    if ( channel4_.get_connection().get_type() == Connection_base::Connection_type::undefined )
    {
        if ( channel4_.get_connection().get_host_name() == szComString ) 
        {
            // Bestimmt die Gueltigkeit der Protokollversion

            // Alle Teilnehmer erhalten unbekanntes Protokoll
            memset(protocolVersion_, 0, sizeof(protocolVersion_));
            // Ggf. unbekannte Dockingstation
            channel4_.get_protocol().reset_timeboy_station();
        }

        // Je nach Typ der Verbindung die Klasse initialisieren.
        switch (iType) 
        {
        case 1: // RS232 direkt (ohne autobaud)
            // Schnittstelle oeffnen
            channel4_.set_auto_baud(false);
            res = channel4_.open_connection(Connection_base::Connection_type::serial, szComString, iComValue, iComTimeOut);
            break;

        case 3: // TCP/IP
            // Schnittstelle oeffnen
            channel4_.set_auto_baud(false);
            res = channel4_.open_connection(Connection_base::Connection_type::socket, szComString, iComValue, iComTimeOut);
            break;

        case 4: // RS232 direkt mit autobaud
            // Schnittstelle oeffnen
            channel4_.set_auto_baud(true);
            res = channel4_.open_connection(Connection_base::Connection_type::serial, szComString, iComValue, iComTimeOut);
            break;

        default: // Unbekannter Schnittstellentyp
            res = IV_ERROR_FAIL;
        }

        // Nach Erfolg die Grundeinstellung wiederherstellen
        if (res == IV_ERROR_SUCCESS) 
        {
            setDeviceSender( deviceAddress );

            if ( channel4_.get_connection().get_host_name() == szComString ) 
            {
                // Bestimmt die Gueltigkeit der Protokollversion

                // Alle Teilnehmer erhalten unbekanntes Protokoll
                memset( protocolVersion_, 0, sizeof(protocolVersion_) );
                // Ggf. unbekannte Dockingstation
                channel4_.get_protocol().reset_timeboy_station();
            }
        }
    } 

    return res;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
void ChannelBridge::close(void)
{
    // Kanal 4 ist Massgebend
    channel4_.close_connection();

    // Bevorzugt AEIII+ pruefen
    preferAEIII_ = false;
}

// ##################################################################
/**
* @brief Ermitteln ob die Schnittstelle geoeffnet ist.
*/
bool ChannelBridge::isOpen()
{
    // Channel4 ist Massgebend!
    return channel4_.is_connection_open();
}

// ##################################################################
/**
* @brief Schnittstelle nach Fehler bereinigen.
*/
void ChannelBridge::resetWithError( int err ) 
{
    // Gehaltene Protokollversion auswerten
    switch( protocolVersion_[(channel4_.get_protocol().get_device_address_receiver() == 0xFE) ? 0 : channel4_.get_protocol().get_device_address_receiver()] ) 
    {
    case 0:
    case 1:
    case 2:
    case 3: // Anstehende Daten verwerfen
        if ( err == IV_ERROR_CRC_PACKET || err == IV_ERROR_CHANNEL_COMMUNICATION || err == IV_ERROR_RECEIVE ) 
        {
            channel4_.get_connection().flush();
        }
    case 4: 
        // Protocolversion zuruecksetzen
        if ( err == IV_ERROR_CHANNEL_COMMUNICATION || err == IV_ERROR_RECEIVE ) 
        {
            setProtocolVersion( deviceAddressReceiver(), 0 );
            // Ggf. unbekannte Dockingstation
            channel4_.get_protocol().reset_timeboy_station();
        }
    }
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
unsigned char ChannelBridge::setDeviceAddressReceiver( unsigned char deviceAddress )
{
    unsigned char deviceAddressPrev = channel4_.get_protocol().get_device_address_receiver();

    channel1_2_.SetDeviceIDRemote( deviceAddress );
    channel4_.get_protocol().set_device_address_receiver( deviceAddress );

    return deviceAddressPrev;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
unsigned char ChannelBridge::deviceAddressReceiver()
{
    return channel4_.get_protocol().get_device_address_receiver();
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
unsigned char ChannelBridge::setDeviceSender( unsigned char deviceAddress )
{
    unsigned char deviceAddressPrev = channel4_.get_protocol().get_device_address_sender();

    channel1_2_.SetDeviceID( deviceAddress );
    channel4_.get_protocol().set_device_address_sender( deviceAddress );

    return deviceAddressPrev;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
unsigned char ChannelBridge::deviceAddressSender()
{
    return channel4_.get_protocol().get_device_address_sender();
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
int ChannelBridge::comReleaseHandle()
{ 
    // CChannel4
    return channel4_.get_connection().detach();
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
int ChannelBridge::comUseHandle( HANDLE hCom )
{
    // CChannel4
    return  channel4_.get_connection().attach( Connection_base::Connection_type::serial, hCom, 800 );
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
HANDLE ChannelBridge::getComPort()
{
    HANDLE hCom = INVALID_HANDLE_VALUE;

    if ( channel4_.get_connection().get_type() == Connection_base::Connection_type::serial )
    {
        hCom = channel4_.get_connection().get_serial_port();
    }
    else if ( channel4_.get_connection().get_type() == Connection_base::Connection_type::socket )
    {
        hCom = (HANDLE)channel4_.get_connection().get_socket();
    }

    return hCom;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
int ChannelBridge::setComPort(HANDLE hCom, int iType, int iComValue, int iTimeOut)
{
    switch (iType)
    {
    case 0:
    case 2:
        // Muss undefiniert sein oder mit gewuenschtem uebereinstimmen
        if ((channel4_.get_connection().get_type() == Connection_base::Connection_type::undefined) ||
            ((channel4_.get_connection().get_type() == Connection_base::Connection_type::serial) && (channel4_.get_connection().get_serial_port() == hCom)) )
        {
            if (channel4_.get_connection().change_state_serial(hCom, iComValue, iTimeOut) != IV_ERROR_SUCCESS)
            {
                return -IV_ERROR_FAIL;
            }

            // CChannel4
            channel4_.get_connection().attach(Connection_base::Connection_type::serial, hCom, iTimeOut);
        }
        else
        {
            return -IV_ERROR_FAIL;
        }
        break;

    case 1:
    case 3:
        // Muss undefiniert sein oder mit gewuenschtem uebereinstimmen
        if ((channel4_.get_connection().get_type() == Connection_base::Connection_type::undefined) ||
            ((channel4_.get_connection().get_type() == Connection_base::Connection_type::socket) && (channel4_.get_connection().get_socket() == (SOCKET)hCom)) )
        {
            // CChannel4
            channel4_.get_connection().attach(Connection_base::Connection_type::socket, hCom, iTimeOut);
        }
        else
        {
            return -IV_ERROR_FAIL;
        }
        break;
    }

    return IV_ERROR_SUCCESS;
}

void ChannelBridge::setCallBackFunction(int (*pFunction)(void))
{
    channel1_2_.SetCBFunction(pFunction);
    channel4_.get_protocol().set_call_back_function( pFunction == nullptr ? std::function<int()>() : pFunction );
}

// Setzen des Logdateinamen.
void ChannelBridge::setLogFileName(const std::string &szLogFileName)
{
    channel4_.set_trace_file_path(szLogFileName);
}	


//#########################################################
//####################### HELPER ##########################
//#########################################################

int ChannelBridge::read(uint8_t *buffer, std::size_t receiveCount, std::size_t &receivedCount)
{
    return channel4_.read_data(buffer, receiveCount, receivedCount);
}

int ChannelBridge::write(const uint8_t *buffer, std::size_t transmitCount, std::size_t &transmittedCount)
{
    return channel4_.write_data(buffer, transmitCount, transmittedCount);
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
unsigned char ChannelBridge::getProtocolVersion(unsigned char deviceAddress)
{
    switch(protocolVersion_[(deviceAddress == 0xFE) ? 0 : deviceAddress]) 
    {
    case 1: // Version 1 des Protokolls
    case 2: // Version 2 des MM-Protokolls
    case 3: // Mehrfachdockingstation
    case 4: // Version 4 des MM-Protokolls
        return protocolVersion_[(deviceAddress == 0xFE) ? 0 : deviceAddress];

    default: // Ermitteln der Protokollversion

        // Bevorzugt auf AEIII+ pruefen?
        if (preferAEIII_ && (channel1_2_.Request((unsigned char)deviceAddress) == TRUE)) 
        { 
            // Pruefen ob Protokollversion 1 vorhanden
            return protocolVersion_[(deviceAddress == 0xFE) ? 0 : deviceAddress] = 1;	
        } 
        else if ( isOpen() && channel4_.get_protocol().is_timeboy_slot(deviceAddress) ) 
        { 
            // Pruefen ob Protokollversion 4 vorhanden
            return protocolVersion_[(deviceAddress == 0xFE) ? 0 : deviceAddress] = 4;
        } 
        else if ( isOpen() && (channel4_.ping(deviceAddress, nullptr, nullptr) >= IV_ERROR_SUCCESS || channel4_.get_last_result() == -IV_ERROR_NEED_USER_PASSWORD) ) 
        { 
            // Pruefen ob Protokollversion 4 vorhanden
            return protocolVersion_[(deviceAddress == 0xFE) ? 0 : deviceAddress] = 4;
        } 
        else if ( isOpen() ) // Schnittstelle kann ggf. aufgrund von Fehlern bereits geschlossen sein.
        {	
            // Handelt es sich ggf. um eine Dockingstation der Version 1
            if ( m_docking_prefer_type == 2 ) 
            { 
                channel4_.get_protocol().set_timeboy_station_enabled( true );
                if ( channel4_.get_protocol().is_timeboy_slot(deviceAddress) )
                    // Pruefen ob Protokollversion 4 vorhanden
                        return protocolVersion_[(deviceAddress == 0xFE) ? 0 : deviceAddress] = 4;

                channel4_.get_protocol().set_timeboy_station_enabled( false );
            }

            // Hoher Modbustimeout aelterer Firmwarestaende uebergehen.
            std::this_thread::sleep_for( std::chrono::milliseconds( 100 ) );

            if (channel1_2_.Request2((unsigned char)deviceAddress) == TRUE) 
            { 
                // Pruefen ob Protokollversion 2 vorhanden
                return protocolVersion_[(deviceAddress == 0xFE) ? 0 : deviceAddress] = 2;
            } 
            else if ( isOpen() ) // Schnittstelle kann ggf. aufgrund von Fehlern bereits geschlossen sein.
            {
                // Hoher Modbustimeout aelterer Firmwarestaende uebergehen.
                std::this_thread::sleep_for( std::chrono::milliseconds( 100 ) );

                if (channel1_2_.Request((unsigned char)deviceAddress) == TRUE) 
                { 
                    // Pruefen ob Protokollversion 1 vorhanden
                    return protocolVersion_[(deviceAddress == 0xFE) ? 0 : deviceAddress] = 1;
                }
            }
        }
    }

    return 0;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
unsigned char ChannelBridge::setProtocolVersion(unsigned char deviceAddress, unsigned char version)
{
    unsigned char old = protocolVersion_[(deviceAddress == 0xFE) ? 0 : deviceAddress];

    protocolVersion_[(deviceAddress == 0xFE) ? 0 : deviceAddress] = version;

    return old;
}

//////////////////////////////////////////////////////////////////////
// Members
//////////////////////////////////////////////////////////////////////

int ChannelBridge::collectData(unsigned char deviceAddress, int *piError)
{
    int iRes = -1;
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    // Loeschen des Datenpuffer
    if (m_record_data[0])
    {
        m_record_size = 0;
        m_record_error = 0;
        memset( m_record_data, 0, sizeof(m_record_data) );
    }

    switch ( getProtocolVersion(deviceAddressReceiver()) )
    {
    case 1:
    case 2:
        // Datensatz ermitteln
        iRes = channel1_2_.ReadRecord((unsigned char)deviceAddressReceiver(), m_record_data, (int*)&m_record_size, &m_record_error);
        break;

    case 4: {
        unsigned char status = 0;
        m_record_size = 0;
        m_record_data[0] = deviceAddressReceiver();

        // Befehlsausfuehrung starten.
        iRes = channel4_.read_record(deviceAddressReceiver(), &m_record_data[3], m_record_size, status);

        iRes = resultToError(iRes, &m_record_error);
        if (iRes)
        {
            // OK
            iRes = status;

            // Groesze anpassen
            m_record_size += 5;
        }
        else
        {
            iRes = -1;
        }
        break;}

    default:
        *piError = IV_ERROR_UNREACHABLE;
        return -1;
    }

    switch ( iRes )
    {
    case 0: // Keine Datensaetze vorhanden		
        m_record_size = 0;
        m_record_error = 0;
        memset( m_record_data, 0, sizeof(m_record_data) );
        return 0;

    case 1: // Online Datensatz
    case 2: // Wiederholter Online Datensatz
        return 2;

    case 3: // Offline Datensatz
    case 4: // Wiederholter Offline Datensatz
        return 1;

    default:
        // Fehler bei Ermittlung
        *piError = m_record_error;
    }

    return -1;
}

int ChannelBridge::getDatensatz( uint8_t *pucBuf, std::size_t &size, int32_t &error)
{
    int iRes = 0;

    // Wenn BusNum des letzten Datensatzpakets 0 ist dann sind 
    // keine gueltigen Daten vorhanden.
    if ( m_record_data[0] == 0 )
    {
        return 0;
    }

    // Daten in Puffer uebergeben
    //unsigned char ucTmp[MAX_MB_PAKET_DATA * 3 + 1];

    // Zu grosze Datensaetze abfangen!
    if ( (m_record_size - 6) > MAX_MB_PAKET_DATA )
    {
        m_record_size -= (m_record_size - 6) - MAX_MB_PAKET_DATA;
    }
    for ( std::size_t i = 0; i < (m_record_size - 6); i++ )
    {
        pucBuf[i] = m_record_data[i + 4];
        //sprintf((char*)&ucTmp[i * 3], "%02X ", (unsigned char)_recordData[i + 4]);
    }
    //ucTmp[i * 3] = 0;

    size = m_record_size - 6;

    //theApp.m_log.Append("\t%s\t%i", ucTmp, *piLength);

    switch( getProtocolVersion( deviceAddressReceiver() ) )
    {
    case 1:
    case 2:
        // Datensatz quittieren (3x -> Versuchen)
        for ( int i = 0; (i < 3) && (iRes = channel1_2_.QuitRecord( deviceAddressReceiver(), m_record_data,
            (int *)&m_record_size, &m_record_error )) == -1; i++ );

            // Quittierung erfolgreich?
            if ( iRes )
            {	
                // Naechsten Datensatz mit Duplikattest anfordern
                for ( int i = 0; (i < 3) && (iRes = channel1_2_.ReadRecord( deviceAddressReceiver(), m_record_data,
                    (int *)&m_record_size, &m_record_error )); i++);
            } 
            break;

    case 4:
    {
        unsigned char status = 0;

        // Befehlsausfuehrung starten.
        iRes = channel4_.quit_record( deviceAddressReceiver(), status );

        iRes = resultToError( iRes, &m_record_error );
        if ( iRes )
        {
            // OK
            // naechsten Datensatz lesen.
            unsigned char status = 0;
            m_record_size = 0;
            m_record_data[0] = deviceAddressReceiver();

            // Befehlsausfuehrung starten.
            iRes = channel4_.read_record( deviceAddressReceiver(), &m_record_data[3], m_record_size, status );

            iRes = resultToError( iRes, &m_record_error );
            if ( iRes )
            {
                // OK
                iRes = status;

                // Groesze anpassen (wegen veralgemeinerter Ausfuehrung am Funktionsanfang)
                m_record_size += 5;
            }
            else
            {
                iRes = -1;
            }
        }
        else
        {
            iRes = -1;
        }
        break;}

    default:
        error = IV_ERROR_UNREACHABLE;
        return -1;
    }

    switch ( iRes )
    {
    case 0: // Keine Datensaetzte vorhanden
        // Aktuell uebergebener ist letzter moeglicher, bei Quittierungsmisserfolg.
        m_record_size = 0;
        m_record_error = 0;
        memset( m_record_data, 0, sizeof( m_record_data ) );
        return 2;

        // Online Datensaetze
    case 1: // Online Datensatz
    case 2: // Wiederholter Online Datensatz
        return 1;

        // Offline Datensaetze
    case 3: // Offline Datensatz
    case 4: // Wiederholter Offline Datensatz
        return 1;

    default:
        error = m_record_error;
    }

    return -1;
}

int ChannelBridge::systemReadRecord(unsigned char deviceAddress, uint8_t *buffer, std::size_t &size, int32_t &error)
{
    int iRes = -1;
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch(getProtocolVersion(deviceAddressReceiver()))
    {
    case 1:
    case 2: 
        error = IV_ERROR_VERSION_FIRMWARE;
        return -1;

    case 4: {
        // OK
        // naechsten Datensatz lesen.
        unsigned char status = 0;
        m_record_size = 0;
        m_record_data[0] = deviceAddressReceiver();

        // Befehlsausfuehrung starten.
        iRes = channel4_.read_system_record(deviceAddressReceiver(), &m_record_data[3], m_record_size, status);

        iRes = resultToError(iRes, &m_record_error);
        if (iRes)
        {
            // OK
            iRes = status;

            // Groesse anpassen
            m_record_size += 5;
        }
        else
        {
            iRes = -1;
        }
        break;}

    default:
        error = IV_ERROR_UNREACHABLE;
        return -1;
    }

    //	TRACE("( %i )\n", iRes);
    switch ( iRes )
    {
    case 0: // Keine Datensaetze vorhanden
        m_record_size = 0;
        m_record_error = 0;
        memset(m_record_data, 0, sizeof(m_record_data));
        return 0;

    case 1: // System Datensatz
        // Daten in Puffer uebergeben, zu grosse Datensaetze sind abfangen!
        for (int i=0; i < (m_record_size - 6); i++)
        {
            buffer[i] = m_record_data[i + 4];
        }
        size = m_record_size - 6;
        return iRes;

    default:
        // Fehler bei Ermittlung
        error = m_record_error;
        resetWithError( error );
        break;
    }

    return -1;
}

int ChannelBridge::systemQuitRecord(unsigned char deviceAddress, int *error)
{
    int iRes = -1;
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1:
    case 2: 
        *error = IV_ERROR_VERSION_FIRMWARE;
        return -1;

    case 4: {
        // OK
        // naechsten Datensatz lesen.
        unsigned char status = 0;

        // Befehlsausfuehrung starten.
        iRes = channel4_.quit_system_record(deviceAddressReceiver(), status);

        iRes = resultToError(iRes, &m_record_error);
        if (iRes)
        {
            // OK
            iRes = status;
        }
        else
        {
            iRes = -1;
        }
        break;}

    default:
        *error = IV_ERROR_UNREACHABLE;
        return -1;
    }

    //	TRACE("( %i )\n", iRes);
    switch ( iRes )
    {
    case 0: // Keine Datensaetze vorhanden
        return 0;

    case 1: // Datensatz quitiert
        return 1;

    default:
        // Fehlercode uebernehmen
        *error = m_record_error;
    }

    return -1;
}

int ChannelBridge::systemRestoreRecords(unsigned char deviceAddress, int *piError)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1:
    case 2: 
        *piError = IV_ERROR_VERSION_FIRMWARE;
        return 0;

    case 4: { // PZEMasterIV mit Protokollversion 4
        int iRes;

        // Befehlsausfuehrung starten.
        iRes = channel4_.restore_system_records(deviceAddressReceiver());

        return resultToError(iRes, piError);}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::readRecord( unsigned char deviceAddress, uint8_t *buffer, std::size_t &size, int32_t &error )
{
    int iRes = -1;
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1:
    case 2:

        // Datensatz ermitteln
        iRes = channel1_2_.ReadRecord( deviceAddressReceiver(), m_record_data, (int *)&m_record_size, &m_record_error );
        break;

    case 4: {
        // OK
        // naechsten Datensatz lesen.
        unsigned char status = 0;
        m_record_size = 0;
        m_record_data[0] = deviceAddressReceiver();

        // Befehlsausfuehrung starten.
        iRes = channel4_.read_record( deviceAddressReceiver(), &m_record_data[3], m_record_size, status );

        iRes = resultToError(iRes, &m_record_error);
        if (iRes)
        {
            // OK
            iRes = status;

            // Groesse anpassen
            m_record_size += 5;
        }
        else
        {
            iRes = -1;
        }
        break;}

    default:
        error = IV_ERROR_UNREACHABLE;
        return -1;
    }

    //	TRACE("( %i )\n", iRes);
    switch ( iRes )
    {
    case 0: // Keine Datensaetze vorhanden
        m_record_size = 0;
        m_record_error = 0;
        memset(m_record_data, 0, sizeof(m_record_data));
        return 0;

    case 1: // Online Datensatz
    case 2: // Wiederholter Online Datensatz
    case 3: // Offline Datensatz
    case 4: // Wiederholter Offline Datensatz

        // Daten in Puffer uebergeben
        int i;
        // Zu grosse Datensaetze abfangen!
        if ((m_record_size - 6) > MAX_MB_PAKET_DATA)
        {
            m_record_size -= (m_record_size - 6) - MAX_MB_PAKET_DATA;
        }
        for (i=0; i < (m_record_size - 6); i++)
        {
            buffer[i] = m_record_data[i + 4];
        }

        size = m_record_size - 6;
        return iRes;

    default:
        // Fehler bei Ermittlung
        error = m_record_error;
        resetWithError(error);
        break;
    }

    return -1;
}

int ChannelBridge::quitRecord(unsigned char deviceAddress, int *error)
{
    int iRes = -1;
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1:
    case 2:

        // Datensatz ermitteln
        iRes = channel1_2_.QuitRecord( deviceAddressReceiver(), m_record_data, (int *)&m_record_size, &m_record_error );
        break;

    case 4: {
        // OK
        // naechsten Datensatz lesen.
        unsigned char status = 0;

        // Befehlsausfuehrung starten.
        iRes = channel4_.quit_record(deviceAddressReceiver(), status);

        iRes = resultToError(iRes, &m_record_error);
        if (iRes)
        {
            // OK
            iRes = status;
        }
        else
        {
            iRes = -1;
        }
        break;}

    default:
        *error = IV_ERROR_UNREACHABLE;
        return -1;
    }

    //	TRACE("( %i )\n", iRes);
    switch ( iRes ) 
    {
    case 0: // Keine Quittung noetig	
        return 0;

    case 1: // Online Datensatzquit.
    case 2: // Wiederholte Online Datensatzquit.
    case 3: // Offline Datensatzquit.
    case 4: // Wiederholte Offline Datensatzquit.
        return iRes;

    default:
        // Fehler bei Ermittlung
        *error = m_record_error;
        resetWithError(*error);
        break;
    }

    return -1;
}

// Loescht alle Datensaetze
int ChannelBridge::clearData(unsigned char deviceAddress, int *piError)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddressReceiver())) 
    {
    case 1:
    case 2:
        return channel1_2_.ClearData((unsigned char)deviceAddressReceiver(), piError);

    case 4: { // PZEMasterIV mit Protokollversion 4
        int iRes;

        // Befehlsausfuehrung starten.
        iRes = channel4_.clear_records(deviceAddressReceiver());

        return resultToError(iRes, piError);}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Stellt den Datensatzzeiger an Ringbeginn. Alle Datensaetze wieder abrufbar.
int ChannelBridge::restoreRecords(unsigned char deviceAddress, int *piError)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddressReceiver())) {
    case 1:
    case 2: 
        *piError = IV_ERROR_VERSION_FIRMWARE;
        return 0;

    case 4: { // PZEMasterIV mit Protokollversion 4
        int iRes;

        // Befehlsausfuehrung starten.
        iRes = channel4_.restore_records(deviceAddressReceiver());

        return resultToError(iRes, piError);}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::getDateTime(unsigned char deviceAddress, unsigned char *pucDateTime, int *piError)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1:
    case 2:

        if(!channel1_2_.ReadMem(MB_GET_TIME, 7, pucDateTime, piError)) 
        {		
            return 0;
        }
        return 1;

    case 4: {
        int iRes = IV_ERROR_SUCCESS;
        char szTime[20];

        // Befehlsausfuehrung starten.
        iRes = channel4_.read_date_time(deviceAddressReceiver(), szTime);

        iRes = resultToError(iRes, piError);
        if (iRes)
        {
            // OK
            pucDateTime[0] = (szTime[0] - 0x30) * 10 + szTime[1] - 0x30;
            pucDateTime[1] = (szTime[2] - 0x30) * 10 + szTime[3] - 0x30;

            pucDateTime[2] = (szTime[5] - 0x30) * 10 + szTime[6] - 0x30;

            pucDateTime[3] = (szTime[8] - 0x30) * 10 + szTime[9] - 0x30;

            pucDateTime[4] = (szTime[11] - 0x30) * 10 + szTime[12] - 0x30;
            pucDateTime[5] = (szTime[14] - 0x30) * 10 + szTime[15] - 0x30;
            pucDateTime[6] = (szTime[17] - 0x30) * 10 + szTime[18] - 0x30;
        }

        return iRes;}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::setDateTime(unsigned char deviceAddress, const unsigned char *pucDateTime, int *piError)
{
    unsigned char ucDateTime[8];

    if (pucDateTime == NULL ||
        (pucDateTime[0] != 20) ||
        (pucDateTime[1] > 99) ||
        (pucDateTime[2] < 1) || (pucDateTime[2] > 12) || 
        (pucDateTime[3] < 1) || (pucDateTime[3] > 31) ||
        (pucDateTime[4] > 23) || (pucDateTime[5] > 59) || (pucDateTime[6] > 59))
    {	
        time_t tm_time;
        struct tm *ptm;
        tm_time = time(NULL);
        ptm = localtime(&tm_time);

        ucDateTime[0] = (ptm->tm_year + 1900) / 100;
        ucDateTime[1] = (ptm->tm_year + 1900) % 100;
        ucDateTime[2] = ptm->tm_mon + 1;
        ucDateTime[3] = ptm->tm_mday;
        ucDateTime[4] = ptm->tm_hour;
        ucDateTime[5] = ptm->tm_min;
        ucDateTime[6] = ptm->tm_sec;
    }
    else
    {
        memcpy(ucDateTime, pucDateTime, 7);
    }

    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    int iRes = IV_ERROR_SUCCESS;
    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1:
    case 2:

        if(!channel1_2_.WriteMem(MB_SET_TIME, 7, ucDateTime, piError))
        {	
            return 0;
        }
        return 1;

    case 4: {
        char szTime[27];

        snprintf(szTime, sizeof( szTime ) - 1, "%02i%02i-%02i-%02i %02i:%02i:%02i", 
            ucDateTime[0], ucDateTime[1], ucDateTime[2], ucDateTime[3], 
            ucDateTime[4], ucDateTime[5], ucDateTime[6]);

        // Befehlsausfuehrung starten.
        iRes = channel4_.write_date_time(deviceAddressReceiver(), szTime);

        return resultToError(iRes, piError);}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::sendInfotext(unsigned char deviceAddress, const std::string &text, int *piError)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1: {
        unsigned char Buffer[245];
        unsigned int counter;
        unsigned int sizeText = static_cast<unsigned int>( text.size() );

        Buffer[0] = 0xFF;
        Buffer[1] = 0;
        Buffer[2] = 0;

        for(counter=0; counter < sizeText; counter++)
        {
            Buffer[counter + 3] = text[counter];
        }

        if(!channel1_2_.WriteMem(INFOTEXT_MEMORY, sizeText + 3, &Buffer[0], piError))
        {
            return 0;
        }

        return 1;}

    case 2:
        *piError = IV_ERROR_FUNCTION_UNAVAILABLE;
        return 0;

    case 4: {
        int iRes = IV_ERROR_SUCCESS;

        // Befehlsausfuehrung starten.
        iRes = channel4_.write_memo_to_display(deviceAddressReceiver(), 0, 0, 0, 0, 0, text);

        return resultToError(iRes, piError);}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::sendMessage(unsigned char deviceAddress, unsigned char ucDelayTime, 
                               unsigned char ucAction,  unsigned char ucAudio, 
                               const std::string &text, int *piError)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1:
    case 2: {
        unsigned char Buffer[245];
        unsigned int counter;
        unsigned int sizeText = static_cast<unsigned int>( text.size() );

        Buffer[0] = ucDelayTime;
        Buffer[1] = ucAction;
        Buffer[2] = ucAudio;

        for(counter=0; (counter < sizeText) && (counter < (int)(sizeof(Buffer) - 3)); counter++)
        {
            Buffer[3 + counter] = text[counter];
        }

        if(!channel1_2_.WriteMem(DISPLAY_MEMORY, counter + 3, &Buffer[0], piError))
        {
            return 0;
        }
        return 1;}

    case 4: {
        int iRes = IV_ERROR_SUCCESS;

        // Befehlsausfuehrung starten.
        iRes = channel4_.write_message_to_display(deviceAddressReceiver(), ucDelayTime, ucAction, ucAudio, text);

        return resultToError(iRes, piError);}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Datensatzbeschreibung oder Listenbeschreibung aus dem Geraet lesen.
int ChannelBridge::getTableDescription(unsigned char deviceAddress, unsigned char ucList, int *piError)
{
    int res = 0;
    // Busnummer uebernehmen
    setDeviceAddressReceiver( deviceAddress );

    // Für den späteren Zugriff auf die Daten des Puffers die ProtocolVersion halten.
    unsigned char protocolVersion = getProtocolVersion( deviceAddress );

    switch ( protocolVersion )
    {
    case 1:  // AEIII+
    case 2:  // PZE - Master IV        
        res = channel1_2_.GetTables( ucList, piError );
        break;

    case 4:
        // Befehlsausfuehrung starten.
        res = resultToError( channel4_.getTablesWithTimeboySlot( deviceAddress, (ucList) ? (char*)"LIST" : (char*)"RECORD" ), piError );
        break;

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    if ( res )
    {
        if ( ucList )
            protocolVersionListDescription_ = protocolVersion;
        else
            protocolVersionRecordDescription_ = protocolVersion;
    }

    return res;
}

// Listenbeschreibung aus dem Geraet lesen.
int ChannelBridge::getListName(unsigned char deviceAddress, unsigned char ucList, std::string &name, int *piError)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1:
    case 2:
        return channel1_2_.GetListName(ucList, name, piError);
        break;

    case 4:
        *piError = IV_ERROR_FUNCTION_UNAVAILABLE;
        return 0;

    default:
        *piError = IV_ERROR_UNREACHABLE;
        return 0;
    }

    return 0;
}

int ChannelBridge::writeStudioFile(unsigned char deviceAddress, const std::string &filePath, int *piError)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddressReceiver())) {
    case 1: 
    case 2:
        return channel1_2_.WriteSetupData(filePath, piError);

    case 4: { // PZEMasterIV mit Protokollversion 4
        int iRes = IV_ERROR_SUCCESS;

        // Befehlsausfuehrung starten.
        iRes = channel4_.writeSetupWithTimeboySlot( deviceAddressReceiver(), filePath );

        return resultToError(iRes, piError);}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::readStudioFile(unsigned char deviceAddress, const std::string &filePath, int *piError)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    int res = 0;
    switch (getProtocolVersion(deviceAddressReceiver())) {
    case 1: 
    case 2:
        return channel1_2_.ReadSetupData(filePath, piError);

    case 4: // PZEMasterIV mit Protokollversion 4
        return resultToError(channel4_.readSetupWithTimeboySlot(deviceAddressReceiver(), filePath), piError);

    default:
        *piError = IV_ERROR_UNREACHABLE;
        res = 0;
    }

    return res;
}

int ChannelBridge::modifyStudioFile(const std::string &sourceFile, const std::string &destFile, const std::string &changeData, int flags, int *piError)
{
    (void)sourceFile;
    (void)destFile;
    (void)changeData;
    (void)flags;
    (void)piError;

#if !defined(DATAFOX_STUDIO_IV)
    CDFFileStudioXML file;
    int fileError = 0;
    struct stat st;

    if (stat(sourceFile.c_str(), &st) != 0)
    {
        // Datei konnte nicht ermittelt werden
        *piError = IV_ERROR_FILE_OPEN;
        return 0;
    }

    // Reset der StudioFile-Klasse
    file.SetDefault();
    file.SetLibraryVersion( libraryVersion );

    // Laden der Dateidaten. Keine Beachtung der Studioversion.
    if (!file.load(sourceFile, &fileError, static_cast<FileStudio::Enum_flag_ignore_version>( FileStudio::fiv_min_studio | FileStudio::fiv_should_studio ))) {
        switch (fileError) {
        case -3:
            *piError = IV_ERROR_FILE_READ;
            break;
        case -7: // Mindestversion DLL wird nicht eingehalten.
            *piError = IV_ERROR_SETUPFILE_MIN_LIBRARY_VERSION;
            break;
        case -8: // Sollteversion DLL
            *piError = IV_ERROR_SETUPFILE_SHOULD_LIBRARY_VERSION;
            break;
        case -9: // Format der Setupdatei
            *piError = IV_ERROR_FILE_FORMAT;
            break;
        case -4: // CRC in Dateidaten
        case -10: case -11: case -12: case -13: case -14: case -15: case -16: case -17: case -18:
            *piError = IV_ERROR_FILE_CRC;
            break;
        default:
            *piError = IV_ERROR_FILE_OPEN;
        }
        return 0;
    }

    // StudioDatei modifizieren
    if (file.Modify(changeData.c_str(), piError))
    {
        // StudioDatei speichern
        if (file.save(destFile, &fileError))
        {
            // Alles erfolgreich durchgefuehrt
            return 1;
        }
        else
        {
            // Datei konnte nicht erstellt werden.
            *piError = IV_ERROR_FILE_OPEN;
        }
    }
#endif
    return 0;
}

int ChannelBridge::checkDevice(unsigned char deviceAddress, int *piError, int iPollRetry)
{
    int iRes = IV_ERROR_SUCCESS;

    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1:
    case 2:
        return channel1_2_.CheckDevice( iPollRetry, piError );

    case 4: { // PZEMasterIV mit Protokollversion 4
        std::string version;

        // Befehlsausfuehrung starten.
        iRes = channel4_.read_firmware_version(deviceAddressReceiver(), version);

        return resultToError(iRes, piError);}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    // Letzter Fehler ist aktueller
    return 0;
}

int ChannelBridge::restart(unsigned char deviceAddress, int *piError)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 2:
        return channel1_2_.Restart( piError );

    case 1:
    case 4:
        *piError = IV_ERROR_VERSION_FIRMWARE;
        break;

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    // Letzter Fehler ist aktueller
    return 0;
}

int ChannelBridge::reset(unsigned char deviceAddress, int mode, int *piError)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    int iRes = IV_ERROR_SUCCESS;
    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1:
    case 2:
        *piError = IV_ERROR_VERSION_FIRMWARE;
        break;

    case 4:
        // Befehlsausfuehrung starten.
        iRes =  channel4_.system_reset(deviceAddress, mode);

        iRes = resultToError(iRes, piError);
        break;

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    // Letzter Fehler ist aktueller
    return iRes;
}

int ChannelBridge::getProtocolVersion2(unsigned char deviceAddress, unsigned char detailed, int *piError)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    unsigned char Buf[4];

    // Protokollverwendung pruefen
    unsigned char pv = getProtocolVersion(deviceAddressReceiver());
    switch(pv)
    {
    case 1:
    case 2:
        // Detailierte Version auslesen.
        if (detailed)
        {
            // Version fuer Setup lesen
            if (!channel1_2_.ReadMem(MB_GET_VERSION, 2, Buf, piError))
            {
                return 0;
            }
        }
        else 
        {
            if (pv == 1)
            {
                // Es wird angenommen ein AEIII+ mit neuester Com-Version
                Buf[0] = 0x04;
                Buf[1] = 0x02;
            }
            else
            {
                // Es wird angenommen ein PZE Master IV mit neuester Com-Version
                Buf[0] = 0x05;
                Buf[1] = 0x01;
            }
        }
        break;

    case 3: 
        // Ab dem Protokolltyp 3 wird nur noch ueber die Firmwareversion entschieden.
        Buf[0] = 0x06;
        Buf[1] = 0x00;
        break;

    case 4:
        // Ab dem Protokolltyp 4 wird nur noch ueber die Firmwareversion entschieden.
        Buf[0] = 0x07;
        Buf[1] = 0x00;
        break;

    default:
        *piError = IV_ERROR_UNREACHABLE;
        return 0;
    }


    // Version fuer Kommunikation

    // Ab Firmware 04.01.00 Protokollanpassung (Version 4) und Listen in Tabellenart abhandeln.
    if ((Buf[0] == 7 && Buf[1] == 0) ||
        // Ab Firmware 04.01.00 Protokollanpassung (Version 3)
            (Buf[0] == 6 && Buf[1] == 0) ||
            // Ab Firmware 04.00.23 Listengroesse auf 256kByte erhoeht.
            (Buf[0] == 5 && Buf[1] == 2) ||
            // Ab Firmware 04.00.21 Listen im Hintergrund laden
            (Buf[0] == 5 && Buf[1] == 1) ||
            // Ab Firmware 04.00.19 des PZEMasterIV (Version 2)
            (Buf[0] == 5 && Buf[1] == 0) ||
            // Ab Firmware 03.01.21 wegen auslesen der Firmwareversion
            (Buf[0] == 4 && Buf[1] == 2) ||
            // Ab Firmware 03.00.12 mit Zutrittskontrolle
            (Buf[0] == 4 && Buf[1] == 1) ||
            // Ab Firmware 3.0 mit 16kByte Flash fuer Setupdaten und 16kByte Extended Flash
            (Buf[0] == 4 && Buf[1] == 0) )
    { 
        // Version als Rueckgabe
        return (Buf[0] << 8) | Buf[1];
    }

    *piError = IV_ERROR_VERSION_FIRMWARE;
    return 0;
}

int ChannelBridge::getSerialNum(unsigned char deviceAddress, int *piSerialNum, int *piError)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1: 
    case 2:
        return channel1_2_.GetSerialNum( piSerialNum, piError );

    case 4: {
        int iRes;
        unsigned int serial;

        // Befehlsausfuehrung starten.
        iRes = channel4_.read_serial_number(deviceAddressReceiver(), serial);

        iRes = resultToError(iRes, piError);
        if (iRes)
        {
            // OK
            *piSerialNum = serial;
        }

        return iRes;}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// ListenBuffer in AE laden
int ChannelBridge::writeListData(unsigned char deviceAddress, int *piError)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1: 
    case 2: 
        return channel1_2_.WriteListData( &channel4_, piError );

    case 4: // PZEMasterIV mit Protokollversion 4

        // Befehlsausfuehrung starten.
        return resultToError(channel4_.writeListWithTimeboySlot(deviceAddressReceiver()), piError);

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// ListenBuffer in AE laden
int ChannelBridge::writeEntranceListData(unsigned char deviceAddress, int iListNum, int *piError)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    // Protokollklasse
    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 4:
        *piError = IV_ERROR_FUNCTION_UNAVAILABLE;
        return 0;

    case 1:
    case 2:
        return channel1_2_.WriteEntranceListData( iListNum, piError );

    default:
        *piError = IV_ERROR_UNREACHABLE;
        return 0;
    }

    return 0;
}

// ListenBuffer in AE laden
int ChannelBridge::writeEntrance2ListData(unsigned char deviceAddress, int iListNum, int *piError)
{
    int iRes = IV_ERROR_SUCCESS;

    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1:
    case 2:
        *piError = IV_ERROR_VERSION_FIRMWARE;
        return 0;

    case 4: 
        // Befehlsausfuehrung starten.
        // Listendaten zum AEIV uebertragen
        return resultToError(channel4_.writeEntrance2ListWithTimeboySlot(deviceAddress, iListNum), piError);

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// ListenBuffer in AE laden
int ChannelBridge::writeTimeboyListData(unsigned char deviceAddress, unsigned char id, unsigned char num, int reserve, int *piError)
{
    int iRes = IV_ERROR_SUCCESS;

    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1:
    case 2:
        *piError = IV_ERROR_VERSION_FIRMWARE;
        return 0;

    case 4:
        // Befehlsausfuehrung starten.
        // Listendaten zum AEIV uebertragen
        return resultToError(channel4_.writeTimeboyListWithTimeboySlot(deviceAddress, id, num, reserve), piError);

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Abfrage ob Datensaetze im AEIII+ vorliegen.
// Erst ab Datensatzweisem lesen.
int ChannelBridge::checkData(unsigned char deviceAddress, int *piError)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1:  // AEIII+
    case 2:  // PZEMasterIV
        return channel1_2_.CheckData(piError);

    case 4: { // PZEMasterIV mit Protokollversion 4
        int iRes = IV_ERROR_SUCCESS;
        unsigned char buf[0x255];
        std::size_t count = 0;
        unsigned char status = 0;

        // Befehlsausfuehrung starten.
        iRes = channel4_.read_record(deviceAddressReceiver(), buf, count, status);

        iRes = resultToError(iRes, piError);
        if (iRes)
        {
            // OK
            iRes = status;
        }
        else
        {
            iRes = -1;
        }
        return iRes;}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return -1;
}

// Schliesst Relais mit Nummer fuer Zeit.
int ChannelBridge::closeRelay(unsigned char deviceAddress, int iNum, int iTime, int *piError)
{
    int iRes = IV_ERROR_SUCCESS;
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1:
    case 2:
        return channel1_2_.CloseRelay((unsigned char)deviceAddressReceiver(), iNum, iTime, piError);

    case 4: { // PZEMasterIV mit Protokollversion 4
//         if (iNum > 4)
//         {
//             *piError = IV_ERROR_FUNCTION_UNAVAILABLE;
//             iRes = 0;
//         }
//         else
//         {
            // Befehlsausfuehrung starten.
            iRes = resultToError(channel4_.write_relay_state(deviceAddressReceiver(), (char*)"INTERN", 0, iNum, 1, iTime), piError);
//         }
        return iRes;}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Ermittelt aktuellen Zustand von Relais mit Nummer fuer Zeit.
int ChannelBridge::getRelayState(unsigned char deviceAddress, int iNum, int *piState, int *piTime, int *piError)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1:
    case 2:
        return channel1_2_.GetRelayState((unsigned char)deviceAddressReceiver(), iNum, piState, piTime, piError);

    case 4: { // PZEMasterIV mit Protokollversion 4
        int iRes = IV_ERROR_SUCCESS;
//         if (iNum > 4)
//         {
//             *piError = IV_ERROR_FUNCTION_UNAVAILABLE;
//             iRes = 0;
//         }
//         else
//         {		
            unsigned char state;
            unsigned short time;

            // Befehlsausfuehrung starten.
            iRes = resultToError(channel4_.read_relay_state(deviceAddressReceiver(), (char*)"INTERN", 0, iNum, state, time), piError);

            if (iRes)
            {
                // OK
                *piState = state;
                *piTime = time;
            }
//        }
        return iRes;}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Oeffnet Relais mit Nummer.
int ChannelBridge::openRelay(unsigned char deviceAddress, int iNum, int *piError)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    int iRes = IV_ERROR_SUCCESS;

    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1:
    case 2:
        return channel1_2_.OpenRelay((unsigned char)deviceAddressReceiver(), iNum, piError);

    case 4: { // PZEMasterIV mit Protokollversion 4
//         if (iNum > 4)
//         {
//             *piError = IV_ERROR_FUNCTION_UNAVAILABLE;
//             iRes = 0;
//         }
//         else
//         {		
            unsigned short time = 0;

            // Befehlsausfuehrung starten.
            iRes = resultToError(channel4_.write_relay_state(deviceAddressReceiver(), (char*)"INTERN", 0, iNum, 0, time), piError);
//         }
        return iRes;}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Daten einer Variablen setzen.
int ChannelBridge::setGlobVar(unsigned char deviceAddress, const char *name, int iType, const char *val, int *piError)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    int iRes = IV_ERROR_SUCCESS;

    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1:
    case 2:
        return channel1_2_.SetGlobVar((unsigned char)deviceAddressReceiver(), name, iType, val, piError);

    case 4: { // PZEMasterIV mit Protokollversion 4

        // Befehlsausfuehrung starten.
        switch (iType)
        {
        case 0: {
            char varName[64];
            memset(varName, 0, sizeof(varName));
            snprintf( varName, sizeof( varName ), "SETUP.%i", *(unsigned char*)name );
            iRes = channel4_.write_variable(deviceAddressReceiver(), varName, val);
            break;}

        case 1: {
            char varName[64];
            memset(varName, 0, sizeof(varName));
            snprintf( varName, sizeof( varName ), "SETUP.%s", name );
            iRes = channel4_.write_variable(deviceAddressReceiver(), varName, val);
            break;}

        case 2:
            iRes = channel4_.write_system_variable(deviceAddressReceiver(), name, val);
            break;
        }

        if (iRes > 0)
        {
            switch(iRes)
            {
            case 1:
            case 2: // Namensbereichsangabe oder Variablenname unbekannt.
                *piError = IV_ERROR_VARIABLE_UNDEFINED;
                iRes = 0;
                break;

            case 3: // Der verwendete Variablenwert ist zu lang.
                *piError = IV_ERROR_VARIABLE_SIZE;
                iRes = 0;
                break;

            case 4: // Der verwendete Variablentype würde nicht eingehalten.
                *piError = IV_ERROR_VARIABLE_TYPE;
                iRes = 0;
                break;

            case 5: // Das Schreib-/Leserecht würde nicht eingehalten.
                *piError = IV_ERROR_VARIABLE_RIGHT;
                iRes = 0;
                break;

            case 6: // Das verwendete Format oder der Wertebereich würde nicht eingehalten.
                *piError = IV_ERROR_VARIABLE_FORMAT_OR_RANGE;
                iRes = 0;
                break;

           default:
                iRes = resultToError(iRes, piError);
            }
        }
        else
        {
            iRes = resultToError(iRes, piError);
        }
        return iRes;}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Daten einer Globalen Variablen lesen.
int ChannelBridge::getGlobVar(unsigned char deviceAddress, const char *name, int iType, char *val, int iValueLength, int *piError)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    // muss mindestens ein Zeichen anfordern.
    if ((iValueLength - 1) <= 0)
    {
        *piError = IV_ERROR_FUNCTION_PARAMETER;
        return 0;
    } 

    int iRes = IV_ERROR_SUCCESS;
    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1:
    case 2:
        return channel1_2_.GetGlobVar((unsigned char)deviceAddressReceiver(), name, iType, val, iValueLength, piError);

    case 4: { // PZEMasterIV mit Protokollversion 4

        std::string lValue;
        // Befehlsausfuehrung starten.
        switch (iType) 
        {
        case 0: {
            char varName[64];
            memset(varName, 0, sizeof(varName));
            snprintf(varName, sizeof( varName ), "SETUP.%i", *(unsigned char*)name);
            iRes = channel4_.read_variable(deviceAddress, varName, lValue);
            break;}

        case 1: {
            char varName[64];
            memset(varName, 0, sizeof(varName));
            snprintf( varName, sizeof( varName ), "SETUP.%s", name );
            iRes = channel4_.read_variable(deviceAddress, varName, lValue);
            break;}

        case 2:
            iRes = channel4_.read_system_variable(deviceAddress, name, lValue);
            break;
        }

        if (iRes > 0)
        {
            switch ( iRes )
            {
            case 1:
            case 2: // Namensbereichsangabe oder Variablenname unbekannt.
                *piError = IV_ERROR_VARIABLE_UNDEFINED;
                iRes = 0;
                break;

            case 3: // Der verwendete Variablenwert ist zu lang.
                *piError = IV_ERROR_VARIABLE_SIZE;
                iRes = 0;
                break;

            case 4: // Der verwendete Variablentype würde nicht eingehalten.
                *piError = IV_ERROR_VARIABLE_TYPE;
                iRes = 0;
                break;

            case 5: // Das Schreib-/Leserecht würde nicht eingehalten.
                *piError = IV_ERROR_VARIABLE_RIGHT;
                iRes = 0;
                break;

            case 6: // Das verwendete Format oder der Wertebereich würde nicht eingehalten.
                *piError = IV_ERROR_VARIABLE_FORMAT_OR_RANGE;
                iRes = 0;
                break;

            default:
                iRes = resultToError(iRes, piError);
            }
        }
        else
        {
            iRes = resultToError(iRes, piError);
        }
        if (iRes) {
            strncpy(val, lValue.c_str(), iValueLength - 1);
        }

        return iRes;}

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::getVersionFirmwareFromFile(const std::string& filePath, std::string& versionFirmware, int *error)
{
    return resultToError(channel4_.getVersionFirmwareFromFile(filePath, versionFirmware), error);
}

int ChannelBridge::writeTextData(unsigned char deviceAddress, const std::string &filePath)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.setTextDataWithTimeboySlot(deviceAddress, filePath);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::writeTouchKeyData(unsigned char deviceAddress, const std::string &filePath)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.setTouchKeyDataWithTimeboySlot(deviceAddress, filePath);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::writeDisplayColors(unsigned char deviceAddress, const std::string &filePath)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.setDisplayColorsWithTimeboySlot(deviceAddress, filePath);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::writeWlanConfigData(unsigned char deviceAddress, const std::string &filePath)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.setWlanConfigDataWithTimeboySlot(deviceAddress, filePath);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::readWlanConfigData(unsigned char deviceAddress, const std::string &filePath)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.getWlanConfigDataWithTimeboySlot(deviceAddress, filePath);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::writeHardwareConfigData(unsigned char deviceAddress, const std::string &filePath)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.setHardwareConfigDataWithTimeboySlot(deviceAddress, filePath);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::readHardwareConfigData(unsigned char deviceAddress, const std::string &filePath)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.getHardwareConfigDataWithTimeboySlot(deviceAddress, filePath);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::writeFileData(unsigned char deviceAddress, const std::string &filePath)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.writeFileDataWithTimeboySlot(deviceAddress, filePath);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::readFileData(unsigned char deviceAddress, const std::string &filePath)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.readFileDataWithTimeboySlot(deviceAddress, filePath);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::writeFileDevice(unsigned char deviceAddress, std::string &filePath, int *piError)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    int iRes = 0;
    switch (getProtocolVersion(deviceAddress)) 
    {
    case 2: // Protokollversion 2 gibt es nur beim PZE-MasterIV
        iRes = channel1_2_.WriteFileDevice(filePath, piError);
        break;

    case 4: // MasterIV mit Protokollversion 4

        // Befehlsausfuehrung starten.
        iRes = resultToError(channel4_.writeDeviceWithTimeboySlot(deviceAddress, filePath), piError);
        break;
    }

    if (iRes)
    {
        // Protokollversion ist aufgrund des Updates auf 0 zu setzen.
        // Naechste Kommunikationsanfage ermittelt die Protokollversion neu.
        protocolVersion_[(deviceAddressReceiver() == 0xFE) ? 0 : deviceAddressReceiver()] = 0;
    }

    return iRes;
}

int ChannelBridge::writeModule(unsigned char deviceAddress, int type, const std::string &filePath, int *piError)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return resultToError(channel4_.writeModuleWithTimeboySlot(deviceAddress, type, filePath), piError);
    }

    *piError = IV_ERROR_UNREACHABLE;    // Zeichensatz- und Grafikdaten
    return 0;
}

//#########################################################
//######## Schnittstelle zu Channel1/Channel2 #############
//#########################################################

int ChannelBridge::getModus(unsigned char deviceAddress, int *error)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 4: return -IV_ERROR_VERSION_FIRMWARE;
    case 2: return channel1_2_.GetModus(error);
    }

    return -IV_ERROR_UNREACHABLE;
}


//#########################################################
//############# Schnittstelle zu Channel4 #################
//#########################################################

void ChannelBridge::setCallBackFunctionWithParams(int (*pFunction)(unsigned int, unsigned int))
{
    channel4_.set_call_back_function_with_parameters( pFunction == nullptr ? std::function<int(unsigned int, unsigned int)>() : pFunction );
}

unsigned int ChannelBridge::crc32(unsigned char *data, unsigned int size, unsigned int crc, bool closure)
{
    return channel4_.get_crc32(data, size, crc, closure);
}

// MODULE_HARDWARE

// Einfache Anfrage ob das Geraet antwortet.
int ChannelBridge::ping(unsigned char deviceAddress)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.ping(deviceAddress, nullptr, nullptr);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeDevice(unsigned char deviceAddress, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_device_type(deviceAddress, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeBoard(unsigned char deviceAddress, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_board_type(deviceAddress, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentRS485_0(unsigned char deviceAddress, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_rs485_communication_component(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeTCPIP(unsigned char deviceAddress, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_network_adapter_type(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeModemMC(unsigned char deviceAddress, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_modem_type(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentRS485_1(unsigned char deviceAddress, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_rs485_access_control_component(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeDisplay(unsigned char deviceAddress, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_display_type(deviceAddress, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeKeypad(unsigned char deviceAddress, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_keypad_type(deviceAddress, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeTransponder(unsigned char deviceAddress, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_transponder_reader_type(deviceAddress, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeMMC(unsigned char deviceAddress, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_mmc_type(deviceAddress, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeRadio(unsigned char deviceAddress, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_radio_type(deviceAddress, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeGPS(unsigned char deviceAddress, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_gps_type(deviceAddress, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentUSB(unsigned char deviceAddress, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_usb_component(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeDCF77(unsigned char deviceAddress, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_DCF77_type(deviceAddress, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentInAnalog(unsigned char deviceAddress, unsigned int &val)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_analog_input_component(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentInDigital(unsigned char deviceAddress, unsigned int &val)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_digital_input_component(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentOutDigital(unsigned char deviceAddress, unsigned int &val)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_digital_output_component(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeFingerprint(unsigned char deviceAddress, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_fingerprint_scanner_type(deviceAddress, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeExternalReader(unsigned char deviceAddress, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_external_transponder_reader_type(deviceAddress, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeTouch(unsigned char deviceAddress, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_touchscreen_type(deviceAddress, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getTypeSensor(unsigned char deviceAddress, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_sensor_type(deviceAddress, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentBattery(unsigned char deviceAddress, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_battery_component(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getComponentCharge(unsigned char deviceAddress, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_battery_charger_component(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionIOModule(unsigned char deviceAddress, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_io_module_option(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionTBAdaption(unsigned char deviceAddress, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_timeboy_adaption_option(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionComServer(unsigned char deviceAddress, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_com_server_option(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionTPWrite(unsigned char deviceAddress, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_transponder_write_option(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionHTTP(unsigned char deviceAddress, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_http_option(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionLanguage(unsigned char deviceAddress, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_localization_option(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionBiokeyUpdate(unsigned char deviceAddress, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_fingerprint_scanner_update_option(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setSetupOptions(unsigned char deviceAddress, unsigned short mask, unsigned short options)
{
    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.write_setup_options(deviceAddress, mask, options);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getSetupOptions(unsigned char deviceAddress, unsigned short &mask, unsigned short &options)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_setup_options(deviceAddress, mask, options);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getPayableOptions(unsigned char deviceAddress, unsigned char &mask, unsigned char &options)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_buy_options(deviceAddress, mask, options);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getHWConfig(unsigned char deviceAddress, unsigned char &val)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_hardware_info_version(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getHardwareInfo(unsigned char deviceAddress, StructHardwareInfo &hwInfo)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.Module_table::read_hardware_info(deviceAddress, hwInfo);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getHardwareInfo( unsigned char deviceAddress, HardwareInfo &hardwareInfo )
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_hardware_info( deviceAddress, hardwareInfo );
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getInfos(StructHardwareInfo hwInfo, std::string &info)
{
    return channel4_.hardware_info_to_string(hwInfo, info);
}

// MODULE_SYSTEM

// Systemreset; Speicherbereiche loeschen
//int systemReset(unsigned char deviceAddress, unsigned short modus);

// Firmwareversion lesen
int ChannelBridge::getVersionFirmware(unsigned char deviceAddress, std::string &version)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: 
        {
            int errorNumber = IV_ERROR_SUCCESS;
            if (channel1_2_.GetVersionFirmware(version, &errorNumber))
                return IV_ERROR_SUCCESS;

            return -errorNumber;
        }
    case 4: return channel4_.read_firmware_version(deviceAddress, version);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getVersionBootloader(unsigned char deviceAddress, std::string &version)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1: return -IV_ERROR_VERSION_FIRMWARE;
    case 2: 
        {
            int errorNumber = IV_ERROR_SUCCESS;
            if (channel1_2_.GetVersionBootloader( version, &errorNumber ))
                return IV_ERROR_SUCCESS;

            return -errorNumber;
        }
    case 4: return channel4_.read_bootloader_version(deviceAddress, version);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getSerialNumber(unsigned char deviceAddress, unsigned int &serial)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_serial_number(deviceAddress, serial);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::openComServerMode(unsigned char deviceAddress, const char *comm, unsigned int baudRate, unsigned char byteSize, unsigned char parity, unsigned char stopBits, unsigned short flags, unsigned short timeoutToClose)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2:return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.enter_com_server_mode(deviceAddress, comm, baudRate, byteSize, parity, stopBits, flags, timeoutToClose);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::closeComServerMode(unsigned char deviceAddress)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    return channel4_.exit_com_server_mode(deviceAddress);
}

// Datensegmentuebertragung beginnen/enden
int ChannelBridge::beginSegment(unsigned char deviceAddress, unsigned short art, unsigned int &address, unsigned int size, unsigned int crc)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.enter_write_data_segment(deviceAddress, art, address, size, crc);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::endSegment(unsigned char deviceAddress, unsigned short art, unsigned int crc)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.exit_write_data_segment(deviceAddress, art, crc);
    }

    return -IV_ERROR_UNREACHABLE;
}


// Uebertragen von Speicherdaten in Datei.
int ChannelBridge::writeMemToFile(unsigned char deviceAddress, int type, int address, int size, const std::string &filePath)
{
    // Remoteadresse einstellen
    setDeviceAddressReceiver(deviceAddress);

    // Aufgrund Protokollversion entscheiden
    // Protokollverwendung pruefen
    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2:
        return channel1_2_.WriteMemToFile(type, address, size, filePath.c_str());

    case 4:
        return channel4_.writeMemToFile(deviceAddress, type, address, size, filePath);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Uebertragen von Speicherdaten aus einer Datei.
int ChannelBridge::writeFileToMem(unsigned char deviceAddress, const std::string &filePath)
{
    // Remoteadresse einstellen
    setDeviceAddressReceiver(deviceAddress);

    // Aufgrund Protokollversion entscheiden
    // Protokollverwendung pruefen
    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: 
        return channel1_2_.WriteFileToMem(filePath);

    case 4:
        return channel4_.writeFileToMem(deviceAddress, filePath);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Flash setzen/lesen (die ersten 16kByte Bootloader koennen nur gelesen werden)
int ChannelBridge::setFlash(unsigned char deviceAddress, unsigned int address, unsigned char *buf, int count)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    int err;
    switch (getProtocolVersion(deviceAddress))
    {
    case 1: return channel1_2_.WriteMem(address, count, buf, &err, 32);
    case 2: return channel1_2_.WriteMem(address, count, buf, &err);
    case 4: return channel4_.write_to_flash(deviceAddress, address, buf, count);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getFlash(unsigned char deviceAddress, unsigned int address, unsigned char *buf, int count)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    int err;
    switch (getProtocolVersion(deviceAddress))
    {
    case 1: return channel1_2_.ReadMem(address, count, buf, &err);
    case 2: return channel1_2_.ReadMem(address, count, buf, &err);
    case 4: return channel4_.read_from_flash(deviceAddress, address, buf, count);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getFlashPage( unsigned char deviceAddress, std::size_t pageNum, std::size_t offs, uint8_t *buffer, std::size_t &count )
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1: 
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_flash_page(deviceAddress, pageNum, offs, buffer, count);
    }

    return -IV_ERROR_UNREACHABLE;
}

// EEPROM setzen/lesen 
int ChannelBridge::setEEPROM(unsigned char deviceAddress, unsigned int address, unsigned char *buf, int count)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.write_to_eeprom(deviceAddress, address, buf, count);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getEEPROM(unsigned char deviceAddress, unsigned int address, unsigned char *buf, int count)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_from_eeprom(deviceAddress, address, buf, count);
    }

    return -IV_ERROR_UNREACHABLE;
}

// RAM setzen/lesen 
int ChannelBridge::setRAM(unsigned char deviceAddress, unsigned int address, unsigned char *buf, int count)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.write_to_ram(deviceAddress, address, buf, count);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getRAM(unsigned char deviceAddress, unsigned int address, unsigned char *buf, int count)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return channel4_.read_from_ram(deviceAddress, address, buf, count);
    }

    return -IV_ERROR_UNREACHABLE;
}

// MMC setzen/lesen 
int ChannelBridge::setMMC(unsigned char deviceAddress, unsigned int address, unsigned char *buf, int count)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1: 
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return channel4_.write_to_mmc(deviceAddress, address, buf, count);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getMMC(unsigned char deviceAddress, unsigned int address, unsigned char *buf, int count)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_from_mmc(deviceAddress, address, buf, count);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Systembios aufrufen/verlassen
int ChannelBridge::enterBios(unsigned char deviceAddress, unsigned char protocolVersion)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (protocolVersion)
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.enter_bios_mode(deviceAddress);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::exitBios(unsigned char deviceAddress)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.exit_bios_mode(deviceAddress);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Interface setzen/lesen (RS232, TCPIP...)
int ChannelBridge::setInterface(unsigned char deviceAddress, unsigned char type)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.write_main_interface_type(deviceAddress, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getInterface(unsigned char deviceAddress, unsigned char &type)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_main_interface_type(deviceAddress, type);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Baud-Rate setzen/lesen
int ChannelBridge::setBaud(unsigned char deviceAddress, unsigned int br)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.write_baud(deviceAddress, br);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getBaud(unsigned char deviceAddress, unsigned int &br)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_baud(deviceAddress, br);
    }

    return -IV_ERROR_UNREACHABLE;
}

// DeviceID setzen/lesen
int ChannelBridge::setDeviceIdBios(unsigned char deviceAddress, unsigned short id)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.write_device_address(deviceAddress, id);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getDeviceIdBios(unsigned char deviceAddress, unsigned short &id)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_device_address(deviceAddress, id);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Bootloader aktivieren
int ChannelBridge::enterBootloader(unsigned char deviceAddress, unsigned char protocolVersion)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (protocolVersion)
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.enter_bootloader_mode(deviceAddress);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Monitor Mode vom XPort
int ChannelBridge::enterXPortMonitor(unsigned char deviceAddress, char *data, uint16_t &count)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.enter_xport_monitor_mode(deviceAddress, data, count);
    }

    //return -IV_ERROR_UNREACHABLE;
    return -IV_ERROR_ENTER_MONITOR;
}

// Monitor Mode vom XPort
int ChannelBridge::enterXPortMonitorExt(unsigned char deviceAddress, char *data, uint16_t &count)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.enter_xport_monitor_mode_ext(deviceAddress, data, count);
    }

    //return -IV_ERROR_UNREACHABLE;
    return -IV_ERROR_ENTER_MONITOR;
}

int ChannelBridge::commandXPort(unsigned char deviceAddress, char *buf, uint16_t &count)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.xport_monitor_mode_command(deviceAddress, buf, count);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::exitXPortMonitor(unsigned char deviceAddress)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.exit_xport_monitor_mode(deviceAddress);
    }

    //return -IV_ERROR_UNREACHABLE;
    return -IV_ERROR_EXIT_MONITOR;
}

// Softwareinformation der Anwendungsfirmware abfragen
int ChannelBridge::getSoftwareInfo(unsigned char deviceAddress, unsigned char *swi, int &infoLength)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.Module_table::read_software_info(deviceAddress, swi, infoLength);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Oeffnen einer Tabelle
int ChannelBridge::tableOpen(unsigned char deviceAddress, const std::string &name, unsigned int options, int &handle, std::string &reserved)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.open_table(deviceAddress, name, options, handle, reserved);
    }

    return -IV_ERROR_UNREACHABLE;
}
// Schliessen der Tabelle
int ChannelBridge::tableClose(unsigned char deviceAddress, int handle)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.close_table(deviceAddress, handle);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Setzen eines Filters auf die Tabelle
int ChannelBridge::tableSetFilter(unsigned char deviceAddress, int handle, const std::string &filter)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.write_table_filter(deviceAddress, handle, filter);
    }

    return -IV_ERROR_UNREACHABLE;
}
// Lesen des Filters auf die Tabelle
int ChannelBridge::tableGetFilter(unsigned char deviceAddress, int handle, std::string &filter)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_table_filter(deviceAddress, handle, filter);
    }

    return -IV_ERROR_UNREACHABLE;
}
// Loeschen des Filters der Tabelle
int ChannelBridge::tableRemoveFilter(unsigned char deviceAddress, int handle)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.remove_table_filter(deviceAddress, handle);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Ermitteln der Anzahl Datensaetze in der Tabelle
int ChannelBridge::tableGetRowCount(unsigned char deviceAddress, int handle, unsigned int &currentCount, unsigned int &unsortedCount, unsigned int &deletedCount)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_table_row_count(deviceAddress, handle, currentCount, unsortedCount, deletedCount);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Ermitteln der Anzahl Datensaetze in der Tabelle
int ChannelBridge::tableGetCurrentRow(unsigned char deviceAddress, int handle, unsigned int &currentRow)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_current_table_row(deviceAddress, handle, currentRow);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Ueberspringen von n Datensaetzen (offset kann negativ sein), ab top/aktuell/bottom
int ChannelBridge::tableSetCurrentRow(unsigned char deviceAddress, int handle, int offset, unsigned short direction)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.write_current_table_row(deviceAddress, handle, offset, direction);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Aendern eines Spaltenwerts, des aktuellen Datensatzes
int ChannelBridge::tableSetCurrentRowData(unsigned char deviceAddress, int handle, unsigned short flags, char separator, const char *data, unsigned short dataSize)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.write_current_table_row_data_ext(deviceAddress, handle, flags, separator, data, dataSize);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Aendern eines Spaltenwerts, des aktuellen Datensatzes
int ChannelBridge::tableSetCurrentColumnData(unsigned char deviceAddress, int handle, unsigned short flags, const char *name, const char *data, unsigned short dataSize)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.write_current_table_column_data_ext(deviceAddress, handle, flags, name, data, dataSize);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Aendern eines Spaltenwerts, des aktuellen Datensatzes
int ChannelBridge::tableSetAllRowsToColumnData(unsigned char deviceAddress, int handle, unsigned short flags, const char *name, const char *data, unsigned short dataSize)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.write_all_table_rows_column_data(deviceAddress, handle, flags, name, data, dataSize);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Abfragen eines Spaltenwerts, des aktuellen Datensatzes
int ChannelBridge::tableGetCurrentRowData(unsigned char deviceAddress, int handle, unsigned short flags, char separator, char *data, unsigned short &dataSize)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_current_table_row_data_ext(deviceAddress, handle, flags, separator, data, dataSize);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Abfragen eines Spaltenwerts, des aktuellen Datensatzes
int ChannelBridge::tableGetCurrentColumnData(unsigned char deviceAddress, int handle, unsigned short flags, const char *name, char *data, unsigned short &dataSize)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_current_table_column_data_ext(deviceAddress, handle, flags, name, data, dataSize);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Aendern eines Spaltenwerts, des aktuellen Datensatzes
int ChannelBridge::tableAppendRowData(unsigned char deviceAddress, int handle, unsigned short flags, char separator, const char *data, unsigned short dataSize)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.append_table_row_data(deviceAddress, handle, flags, separator, data, dataSize);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Loescht den aktuellen Datensatz
int ChannelBridge::tableDeleteCurrentRow(unsigned char deviceAddress, int handle, unsigned short flags)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.delete_current_table_row(deviceAddress, handle, flags);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Loescht dee aktuellen Datensaetze, welche durch das gesetzte Filterkriterium verfuegbar sind.
int ChannelBridge::tableDeleteAvailableRows(unsigned char deviceAddress, int handle, unsigned short flags)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.delete_filtrated_table_rows(deviceAddress, handle, flags);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Abragen eines Modulinfos
int ChannelBridge::getInfo(unsigned char deviceAddress, const std::string &name, const std::string &params, std::string &info)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_hardware_module_info(deviceAddress, name, params, info);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Zutrittsbuchung durchfuehren
int ChannelBridge::entrance2Identification(unsigned char deviceAddress, const char *TM, const char *Id, const char *Pin, char *status, int &statusLength)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.access_control_identification(deviceAddress, TM, Id, Pin, status, statusLength);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::entrance2OnlineAction(unsigned char deviceAddress, const char *TM, unsigned int mask, unsigned char type, unsigned char duration)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.access_control_online_action(deviceAddress, TM, mask, type, duration);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setNotifyOnDevice(unsigned char deviceAddress, int type, int state)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.write_notify(deviceAddress, type, state);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Datensatz an Fingerprint-Tabelle anfuegen.
int ChannelBridge::fingerprintAppendRecord(unsigned char deviceAddress, unsigned char type, const unsigned char *fingertemplate, int size)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.append_fingerprint_record(deviceAddress, type, fingertemplate, size);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Datensatz aus Fingerprint-Tabelle lesen.
int ChannelBridge::fingerprintGetRecord(unsigned char deviceAddress, unsigned char &type, int pid, int fid, unsigned char *fingertemplate, int &size)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_fingerprint_record(deviceAddress, type, pid, fid, fingertemplate, size);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Datensatz aus Fingerprint-Tabelle loeschen.
int ChannelBridge::fingerprintDeleteRecord(unsigned char deviceAddress, int pid, int fid)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.delete_fingerprint_record(deviceAddress, pid, fid);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Liste mit Fingerprintmoduldaten erzeugen.
int ChannelBridge::fingerprintList(unsigned char deviceAddress, int flags, unsigned int &recordCount)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.generate_fingerprint_id_list(deviceAddress, flags, recordCount);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Backup der Fingerprints erstellen
int ChannelBridge::fingerprintBackup(unsigned char deviceAddress, const std::string &filePath, int flags)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.fingerprintBackup(deviceAddress, filePath, flags);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Flags:
// Bit 1: gesetzt bedeutet: Enthaltene Fingertemplates beibehalten.
// Wiederherstellen der Fingerprints
int ChannelBridge::fingerprintRestore(unsigned char deviceAddress, const std::string &filePath, int flags)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.fingerprintRestore(deviceAddress, filePath, flags);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Baud-Rate setzen/lesen
int ChannelBridge::setSystemFlags(unsigned char deviceAddress, unsigned int flags)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return channel4_.write_system_flags(deviceAddress, flags);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getSystemFlags(unsigned char deviceAddress, unsigned int &flags)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return channel4_.read_system_flags(deviceAddress, flags);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setOptionFirmware(unsigned char deviceAddress, int mask, int options)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return channel4_.setOptionFirmware(deviceAddress, mask, options);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::getOptionFirmware(unsigned char deviceAddress, int &mask, int &options)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return channel4_.getOptionFirmware(deviceAddress, mask, options);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Setzt den zu verwendenden Zeichensatztyp fuer sendMessage oder Sendinfotext
int ChannelBridge::setTypeFont(unsigned char deviceAddress, int iType, int *piError)
{
    int iRes = IV_ERROR_SUCCESS;
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1:
    case 2:
        *piError = IV_ERROR_VERSION_FIRMWARE;
        break;

    case 4: // PZEMasterIV mit Protokollversion 4
        // Befehlsausfuehrung starten.
        return resultToError(channel4_.write_font_type(deviceAddressReceiver(), iType), piError);

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

// Passwort pruefen
int ChannelBridge::passwordCheck(unsigned char deviceAddress, char *passwort)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.write_setup_password(deviceAddress, passwort);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Abragen eines Modulinfos
int ChannelBridge::getPasswordKey(unsigned char deviceAddress, char *key)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_setup_password_key(deviceAddress, key);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Tastendruck ueber Kommunikationsbefehl simulieren
int ChannelBridge::pressVirtualKey(unsigned char deviceAddress, unsigned char vk, unsigned char flags, int *piError)
{
    int iRes = IV_ERROR_SUCCESS;
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1:
    case 2:
        *piError = IV_ERROR_VERSION_FIRMWARE;
        break;

    case 4: // PZEMasterIV mit Protokollversion 4
        // Befehlsausfuehrung starten.
        return resultToError(channel4_.write_virtual_key(deviceAddressReceiver(), vk, flags), piError);

    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    return 0;
}

int ChannelBridge::getFlashStatus(unsigned char deviceAddress, int *piStatus, int *piError)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    unsigned int serial = 0;

    int iRes = 0;
    switch (getProtocolVersion(deviceAddressReceiver()))
    {
    case 1:
    case 2:
        *piError = IV_ERROR_VERSION_FIRMWARE;
        break;

    case 4:

        iRes = resultToError(channel4_.read_serial_number(deviceAddressReceiver(), serial), piError);
        if (iRes)
        {
            // Flags in Statusvariable uebernehmen
            *piStatus = (channel4_.get_protocol().get_last_flags() & 0x70) >> 4;
        }
        break;
    default:
        *piError = IV_ERROR_UNREACHABLE;
    }

    // Letzter Fehler ist aktueller
    return iRes;
}

// Passwort fuer die verschluesselte Kommunikation
int ChannelBridge::setCommunicationPassword(const char *password, uint8_t size, uint32_t sessionTimeout, bool global)
{
    // Aufruf an CChannel4 weiterleiten
    return channel4_.get_protocol().set_communication_password(password, size, sessionTimeout, global);
}

// Pruefen ob eine Globale Passwortvergabe moeglich ist
int ChannelBridge::isGlobalCommunicationPassword()
{
    // Aufruf an CChannel4 weiterleiten
    return channel4_.get_protocol().aes_crypt_session().is_global_communication_password();
}

int ChannelBridge::recordVolume(unsigned char deviceAddress, int &number, int &memoryUsage, int &memorySize)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.read_records_volume(deviceAddress, number, memoryUsage, memorySize);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Funktionen für Pufferzugriff
int ChannelBridge::GetTableCount(unsigned char ucList)
{
    // Anhand der zuvor angesprochenen Protocolversion bei getTableDescription den weiteren Zugriff steuern.
    unsigned char protocolVersion = (ucList) ? protocolVersionListDescription_ : protocolVersionRecordDescription_;

    switch ( protocolVersion )
    {
    case 1:
    case 2:
        return channel1_2_.GetTableCount( ucList );

    case 4:
        return channel4_.GetTableCount( ucList );
    }

    return 0;
}

int ChannelBridge::GetTableRecord(unsigned char ucNum, std::string& szName, unsigned char *pucFieldCount, unsigned char *pucVariable, unsigned char ucList)
{
    // Anhand der zuvor angesprochenen Protocolversion bei getTableDescription den weiteren Zugriff steuern.
    unsigned char protocolVersion = (ucList) ? protocolVersionListDescription_ : protocolVersionRecordDescription_;

    // Bei Listen kann es nötig sein auf die Zutrittslistendaten zu verzweigen.
    if ( ucList && ucNum >= 100 )
    {
        // Auf Zutrtittskontrolllisten verzweigen
        if (ucNum >= 100 && ucNum < 104) {
            // Version 1
            *pucVariable = 0;
            return channel1_2_.GetTableRecordEntrance(ucNum, szName, pucFieldCount);
        } else if (ucNum >= 200 && ucNum < 208) {
            // Version 2
            *pucVariable = 0;
            return channel4_.GetTableRecordEntrance2(ucNum, szName, pucFieldCount);
        }
    }

    switch ( protocolVersion )
    {
    case 1:
    case 2:
        return channel1_2_.GetTableRecord( ucNum, szName, pucFieldCount, pucVariable, ucList );

    case 4:
        return channel4_.GetTableRecord( ucNum, szName, pucFieldCount, pucVariable, ucList );
    }

    return 0;
}

int ChannelBridge::GetTableField(unsigned char ucTableNum, unsigned char ucFieldNum, std::string& szName, unsigned char *pucType, unsigned char *pucLength, unsigned char ucList)
{
    // Anhand der zuvor angesprochenen Protocolversion bei getTableDescription den weiteren Zugriff steuern.
    unsigned char protocolVersion = (ucList) ? protocolVersionListDescription_ : protocolVersionRecordDescription_;

    // Bei Listen kann es nötig sein auf die Zutrittslistendaten zu verzweigen.
    if ( ucList && ucTableNum >= 100 )
    {
        // Auf Zutrittskontrolllisten verzweigen
        if (ucTableNum >= 100 && ucTableNum < 104) {
            // Version 1
            return channel1_2_.GetTableFieldEntrance(ucTableNum, ucFieldNum, szName, pucType, pucLength);
        } else if (ucTableNum >= 200 && ucTableNum < 208) {
            // Version 2
            return channel4_.GetTableFieldEntrance2(ucTableNum, ucFieldNum, szName, pucType, pucLength);
        }
    }

    switch ( protocolVersion )
    {
    case 1:
    case 2:
        return channel1_2_.GetTableField( ucTableNum, ucFieldNum, szName, pucType, pucLength, ucList );

    case 4:
        return channel4_.GetTableField( ucTableNum, ucFieldNum, szName, pucType, pucLength, ucList );
    }

    return 0;
}

void ChannelBridge::ClearListBuffer()
{
    // Importiert wird nur in Channel4, bei Load wird ggf. von Channel1_2 auf Channel4Buffer zugegriffen.
    channel4_.ClearListBuffer();
}

int ChannelBridge::MakeListe(int ListNum, int ListCnt, int ListSize, const unsigned char *Buf)
{
    // Importiert wird nur in Channel4, bei Load wird ggf. von Channel1_2 auf Channel4Buffer zugegriffen.
    return channel4_.MakeListe( ListNum, ListCnt, ListSize, Buf );
}

void ChannelBridge::ClearEntranceListBuffer( int iListNum )
{
    channel1_2_.ClearEntranceListBuffer(iListNum);
}

int ChannelBridge::MakeEntranceList( int ListNum, int ListCnt, int ListSize, const unsigned char *Buf )
{
    return channel1_2_.MakeEntranceList( ListNum, ListCnt, ListSize, Buf );
}

void ChannelBridge::ClearEntrance2ListBuffer( int iListNum )
{
    channel4_.ClearEntrance2ListBuffer( iListNum );
}

int ChannelBridge::MakeEntrance2List( int ListNum, int ListCnt, int ListSize, const unsigned char *Buf )
{
    return channel4_.MakeEntrance2List( ListNum, ListCnt, ListSize, Buf );
}

void ChannelBridge::ClearTimeboyListBuffer( unsigned char id, unsigned char num )
{
    channel4_.ClearTimeboyListBuffer( id, num );
}

int ChannelBridge::MakeTimeboyList( unsigned char id, unsigned char num, unsigned short recordCount, const unsigned char *buf, unsigned int size, int flags, int *error )
{
    return channel4_.MakeTimeboyList( id, num, recordCount, buf, size, flags, error );
}

bool ChannelBridge::WriteListToBinFile( unsigned char listNum, const char *filePath )
{
    return channel4_.WriteListToBinFile( listNum, filePath );
}

bool ChannelBridge::WriteListToTextFile( unsigned char listNum, const char *filePath )
{
    return channel4_.WriteListToTextFile( listNum, filePath );
}

bool ChannelBridge::WriteListEntrance2ToBinFile( unsigned char listNum, const char *filePath )
{
    return channel4_.WriteListEntrance2ToBinFile( listNum, filePath );
}

bool ChannelBridge::WriteListEntrance2ToTextFile( unsigned char listNum, const char *filePath )
{
    return channel4_.WriteListEntrance2ToTextFile( listNum, filePath );
}

bool ChannelBridge::WriteListTimeboyToBinFile( unsigned char listId, unsigned char listNum, const char *filePath )
{
    return channel4_.WriteListTimeboyToBinFile( listId, listNum, filePath );
}

std::string ChannelBridge::GetFileDeviceFileName()
{
    return channel4_.m_fileDevice.m_fileName;
}

std::string ChannelBridge::GetFileDeviceError()
{
    return channel4_.m_fileDevice.m_errorExtended;
}

std::string ChannelBridge::GetFileDeviceInfo()
{
    return channel4_.m_hardware_info_as_text;
}

FileDevice &ChannelBridge::GetFileDevice()
{
    return channel4_.m_fileDevice;
}

#if !defined( DATAFOX_STUDIO_IV ) && !defined( DF_API_PRODUCTION )
// Funktions-Stumps

// MODULE_HARDWARE

int ChannelBridge::setTypeDevice(unsigned char deviceAddress, unsigned char type)
{
    (void)deviceAddress;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeBoard(unsigned char deviceAddress, unsigned char type)
{
    (void)deviceAddress;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeTCPIP(unsigned char deviceAddress, unsigned char val)
{
    (void)deviceAddress;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeModemMC(unsigned char deviceAddress, unsigned char val)
{
    (void)deviceAddress;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeDisplay(unsigned char deviceAddress, unsigned char type)
{
    (void)deviceAddress;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeKeypad(unsigned char deviceAddress, unsigned char type)
{
    (void)deviceAddress;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeTransponder(unsigned char deviceAddress, unsigned char type)
{
    (void)deviceAddress;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeMMC(unsigned char deviceAddress, unsigned char type)
{
    (void)deviceAddress;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeRadio(unsigned char deviceAddress, unsigned char type)
{
    (void)deviceAddress;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeGPS(unsigned char deviceAddress, unsigned char type)
{
    (void)deviceAddress;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeDCF77(unsigned char deviceAddress, unsigned char type)
{
    (void)deviceAddress;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeFingerprint(unsigned char deviceAddress, unsigned char type)
{
    (void)deviceAddress;
    (void)type;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setTypeExternalReader(unsigned char deviceAddress, unsigned char val)  
{ 
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return channel4_.write_external_transponder_reader_type(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setComponentRS485_0(unsigned char deviceAddress, unsigned char val)
{
    (void)deviceAddress;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setComponentRS485_1(unsigned char deviceAddress, unsigned char val)
{
    (void)deviceAddress;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setComponentUSB(unsigned char deviceAddress, unsigned char val)
{
    (void)deviceAddress;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setComponentInAnalog(unsigned char deviceAddress, unsigned int val)
{
    (void)deviceAddress;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setComponentInDigital(unsigned char deviceAddress, unsigned int val)
{
    (void)deviceAddress;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setComponentOutDigital(unsigned char deviceAddress, unsigned int val)
{
    (void)deviceAddress;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setOptionIOModule(unsigned char deviceAddress, unsigned char val)      
{ 
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return channel4_.write_io_module_option(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setOptionTBAdaption(unsigned char deviceAddress, unsigned char val)    
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return channel4_.write_timeboy_adaption_option(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;

}

int ChannelBridge::setOptionComServer(unsigned char deviceAddress, unsigned char val)     
{ 
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return channel4_.write_com_server_option(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setOptionTPWrite(unsigned char deviceAddress, unsigned char val)       
{ 
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return channel4_.write_transponder_write_option(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setOptionHTTP(unsigned char deviceAddress, unsigned char val)          
{ 
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return channel4_.write_http_option(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setOptionLanguage(unsigned char deviceAddress, unsigned char val)
{
    (void)deviceAddress;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setOptionBiokeyUpdate(unsigned char deviceAddress, unsigned char val)  
{ 
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE; 
    case 4: return channel4_.write_fingerprint_scanner_update_option(deviceAddress, val);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::setPayableOptions(unsigned char deviceAddress, unsigned char mask, unsigned char options)
{
    (void)deviceAddress;
    (void)mask;
    (void)options;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

int ChannelBridge::setHWConfig(unsigned char deviceAddress, unsigned char val)
{
    (void)deviceAddress;
    (void)val;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

// MODULE_SYSTEM

int ChannelBridge::setSerialNumber(unsigned char deviceAddress, unsigned int serial)
{
    (void)deviceAddress;
    (void)serial;
    return -IV_ERROR_FUNCTION_UNAVAILABLE;
}

#endif

// Zutrittsbuchung durchfuehren
int ChannelBridge::accessControlIdentification(unsigned char deviceAddress, int masterId, const char *TM, const char *Id, const char *Pin, char *status, int &statusLength)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.access_control_identification_ext(deviceAddress, masterId, TM, Id, Pin, status, statusLength);
    }

    return -IV_ERROR_UNREACHABLE;
}

int ChannelBridge::accessControlOnlineAction(unsigned char deviceAddress, int masterId, const char *TM, unsigned int mask, unsigned char type, unsigned char duration)
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.access_control_online_action_ext(deviceAddress, masterId, TM, mask, type, duration);
    }

    return -IV_ERROR_UNREACHABLE;
}

// Zutrittskontroll-Befehl für Schließzylinder
int ChannelBridge::accessControlKnobCommand( unsigned char deviceAddress, unsigned char knobCommand, unsigned short masterId, unsigned char knobId, const char *params, unsigned char *data, int *sizeData )
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.access_control_knob_command( deviceAddress,knobCommand, masterId, knobId, params, data, sizeData );
    }

    return -IV_ERROR_UNREACHABLE;
}


// Benutzerdefinierter Befehl
int ChannelBridge::customCommand( unsigned char deviceAddress, unsigned int idendification, unsigned int customCommand, const char *dataIn, int sizeIn, char *dataOut, int *sizeOut )
{
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: return -IV_ERROR_VERSION_FIRMWARE;
    case 4: return channel4_.custom_command( deviceAddress, idendification, customCommand, dataIn, sizeIn, dataOut, sizeOut );
    }

    return -IV_ERROR_UNREACHABLE;
}


int ChannelBridge::startTransparencyMode( unsigned char deviceAddress, unsigned short id, unsigned short flags, unsigned int sizeReadBuffer, unsigned int sizeWriteBuffer, int timeout, int *errorNumber )  
{ 
    // Busnummer uebernehmen
    setDeviceAddressReceiver(deviceAddress);
    int iRes = 0;

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: 
        {
            *errorNumber = IV_ERROR_VERSION_FIRMWARE;
            break;
        }
    case 4: 
        {
            iRes = channel4_.enter_transparency_mode( deviceAddress, id, flags, sizeReadBuffer, sizeWriteBuffer, timeout );
            if ( iRes != IV_ERROR_SUCCESS )
            {
                // Spezielle Funktionsfehler auswerten, ansonsten nach Standard auflösen.
                switch ( (iRes >> 8) & 0xFF )
                {
                case 1:
                    *errorNumber = IV_ERROR_ALLOC_MEM;
                    iRes = 0;
                    break;

                case 2:
                    *errorNumber = IV_ERROR_TIMEBOY_IV;
                    iRes = 0;
                    break;

                default:
                    iRes = resultToError( iRes, errorNumber );
                }
            }
            else
            {
                iRes = 1;
            }
            break;
        }

    default:
        *errorNumber = IV_ERROR_UNREACHABLE;
        break;
    }

    return iRes;
}

int ChannelBridge::stopTransparencyMode( unsigned char deviceAddress, int *errorNumber )  
{ 
    int iRes = 0;

    switch (getProtocolVersion(deviceAddress))
    {
    case 1:
    case 2: 
        {
            *errorNumber = IV_ERROR_VERSION_FIRMWARE;
            break;
        }
    case 4: 
        {
            iRes = channel4_.exit_transparency_mode( deviceAddress );
            if ( iRes != IV_ERROR_SUCCESS )
            {
                iRes = resultToError( iRes, errorNumber );
            }
            else
            {
                iRes = 1;
            }
            break;
        }

    default:
        *errorNumber = IV_ERROR_UNREACHABLE;
        break;
    }

    return iRes;
}

int ChannelBridge::isTransparencyMode()  
{ 
    return channel4_.is_transparency_mode();
}

int ChannelBridge::ping( unsigned char deviceAddress, int attempts, int *errorNumber )  
{ 
    // Einen Versuch mehr.
    attempts++;

    for ( int i = 1; i < attempts; ++i )
    {
        if ( channel4_.ping( deviceAddress, nullptr, nullptr ) == IV_ERROR_SUCCESS )
            return i;
    }

    *errorNumber = IV_ERROR_UNREACHABLE;
    return 0;
}

