/**
* @file  file_display_fonts.cpp
* @brief Implementierung der Klasse FileDisplayFonts.
*/
//////////////////////////////////////////////////////////////////////

#include "file/file_display_fonts.h"
#include "file/interface/device_text_structs.h"

#include <memory.h>
#include <stdio.h>
#include <fstream>

FileDisplayFonts::FileDisplayFonts()
{
    // Der Standardmäßig zu verarbeitende Datensatztyp
    setRecordType( Record_type::display_font );

    // Initalisieren
    resetData();

    // Fehlerobjekt zuruecksetzen
    resetError();
}

FileDisplayFonts::~FileDisplayFonts()
{
}

void FileDisplayFonts::resetData()
{
    // Basis Reset
    FileIntelHex::resetData();

    // Hilfsstrukturen aufsetzen
    m_pFontNode = nullptr;
    for ( auto &font : m_pFont )
    {
        font = nullptr;
    }
}

void FileDisplayFonts::Default( unsigned char count )
{
    // Speicher fuer die Daten.
    m_buffer.resize( sizeof( FONTNODE ) + count * sizeof( unsigned int ) + count * sizeof( FONT ) );

    // Hilfsstrukturzeiger
    m_pFontNode = (FONTNODE *)&m_buffer[0];

    // 1.
    // Zeichensaetze init
    // Grundstruktur aufsetzen
    m_pFontNode->version = 1;
    m_pFontNode->cnt     = count;
    m_pFontNode->temp[0] = 0;
    m_pFontNode->temp[1] = 0;
    m_pFontNode->ptrFontArray = sizeof( FONTNODE );

    // Zeigerarray und Zeichensatzstrukturzeiger initialisieren
    for ( unsigned char i = 0; i < m_pFontNode->cnt; ++i )
    {
        // Zeigereintrag auf Zeichensatzstruktur
        *(unsigned int *)&m_buffer[m_pFontNode->ptrFontArray + i * sizeof( unsigned int )] = 
            m_pFontNode->ptrFontArray + m_pFontNode->cnt * sizeof( unsigned int ) + i * sizeof( FONT );

        m_pFont[i] = (FONT *)&m_buffer[m_pFontNode->ptrFontArray + m_pFontNode->cnt * sizeof( unsigned int ) + i * sizeof( FONT )];
    }

}

void FileDisplayFonts::setSizeData( std::size_t size )
{
    m_buffer.resize( size );
    // Reinitialisierung der Strukturzeiger.
    // Hilfsstrukturzeiger
    m_pFontNode = (FONTNODE *)&m_buffer[0];
    // Zeigerarray und Zeichensatzstrukturzeiger initialisieren
    for ( int i = 0; i < m_pFontNode->cnt; ++i )
    {
        m_pFont[i] = (FONT *)&m_buffer[m_pFontNode->ptrFontArray + m_pFontNode->cnt * sizeof( unsigned int ) + i * sizeof( FONT )];
    }
}

void FileDisplayFonts::FontCorrect( int iNum, bool onlyDigits ) 
{
    int j = iNum;

    // Leerzeichen ist so breit wie Doppelpunkt
    m_buffer[m_pFont[j]->ptrCharacter + (m_buffer[m_pFont[j]->ptrTable + (unsigned char)' '] - 1) * m_pFont[j]->bytesPerChar] =
        m_buffer[m_pFont[j]->ptrCharacter + (m_buffer[m_pFont[j]->ptrTable + (unsigned char)':'] - 1) * m_pFont[j]->bytesPerChar];

    // Bei nur Ziffern Zeichensatz
    if ( onlyDigits )
    {
        char q;
        unsigned char width = 0;

        // Suche breitestes Zeichen
        for ( q = '0'; q <= '9'; ++q )
        {
            if( m_buffer[m_pFont[j]->ptrCharacter + (m_buffer[m_pFont[j]->ptrTable + (unsigned char)q] - 1) * m_pFont[j]->bytesPerChar] > width )
            {
                width = m_buffer[m_pFont[j]->ptrCharacter + (m_buffer[m_pFont[j]->ptrTable + (unsigned char)q] - 1) * m_pFont[j]->bytesPerChar];
            }
        }
        // Alle Zeichen so breit wie breitestes
        for ( q = '0'; q <= '9'; ++q )
        {
            m_buffer[m_pFont[j]->ptrCharacter + (m_buffer[m_pFont[j]->ptrTable + (unsigned char)q] - 1) * m_pFont[j]->bytesPerChar] = width;
        }
    }

    // Neuberechnung des 32bit CRC ueber Tabelle und Zeichensatzdaten. (muessen immer direkt hintereinander im Speicher liegen)
    m_pFont[j]->crc  = crc32( &m_buffer[m_pFont[j]->ptrTable], 0x100 + m_pFont[j]->size );
}

bool FileDisplayFonts::load( const std::string &filePath )
{
    // Lesen der Hexdatei
    if ( FileIntelHex::load( filePath ) )
    {
        // Member durch gelesene Daten initialisieren und pruefen
        if ( m_buffer.size() )
        {
            int i;

            // Hilfsstrukturzeiger
            m_pFontNode = (FONTNODE *)&m_buffer[0];

            // 1.
            // Zeichensaetze pruefen

            // Zeigerarray und Zeichensatzstrukturzeiger initialisieren
            if ( m_pFontNode->cnt > max_fonts )
            {
                setErrorState( Error_state::format );
                setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            }
            else
            {
                for ( i = 0; i < m_pFontNode->cnt; ++i )
                {
                    m_pFont[i] = (FONT *)&m_buffer[m_pFontNode->ptrFontArray + m_pFontNode->cnt * sizeof( unsigned int ) + i * sizeof( FONT )];
                }

                // CRC32 pruefen
                for ( i = 0; i < m_pFontNode->cnt; ++i )
                {
                    if ( m_pFont[i]->crc != crc32( &m_buffer[m_pFont[i]->ptrTable], 0x100 + m_pFont[i]->size ) )
                    {
                        setErrorState( Error_state::crc_fail );
                        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                        break;
                    }
                }

            }
        }
        else
        {
            // Fehler zuweisen
            setErrorState( Error_state::no_data );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        }
    }

    return (errorState() == Error_state::none);
}

bool FileDisplayFonts::save( const std::string &filePath, bool appendToFile /*= false*/, bool appendEndLine /*= true */ ) 
{
    bool res = FileIntelHex::save( filePath, appendToFile, appendEndLine );

    if ( res )
    {
        // Die Datei wird fuer einen eventuellen Import in den Firmwarecode als Binaerdatei gespeichert.
        std::string fName( filePath );
        fName.append( ".font.bin" );
        std::ofstream binOut( fName, std::ios::binary );

        // Datei als Binär rausschreiben
        if ( binOut.is_open() == true )
        {
            binOut.write( (const char *)m_buffer.data(), m_buffer.size() );
            binOut.close();
        }
    }

    return res;
}


// Textdaten als letzten Zeichensatz anfuegen.
int FileDisplayFonts::AppendSegmentText(unsigned char *buf, unsigned int size) 
{
    (void)size;
    unsigned char *memFont;
    // Verwaltungsstrukturen fuer Zeichensatz
    FONTNODE *pFontNode;
    std::array<FONT *, max_fonts> pFont;
    TEXTDATA *pText = (TEXTDATA *)buf; // Ist TEXTDATA hat aber den gleichen Aufbau wie FONT
    unsigned int memFontSize;
    int i = 0;

    // Sind bereits Zeichensaetze angegeben?
    if ( m_buffer.size() )
    {
        //###################################
        // Als letzten Zeichensatz anfuegen

        // Speichergroesse von Zeichensaetzen berechnen
        memFontSize = m_pFont[m_pFontNode->cnt - 1]->ptrCharacter + m_pFont[m_pFontNode->cnt - 1]->size;
        // Kopie der Bereits vorhandenen Zeichensatzdaten erstellen
        memFont = new unsigned char[memFontSize];
        // Speicher allokation nicht erfolgreich?
        if ( memFont == nullptr )
        {
            setErrorState( Error_state::alloc_mem );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            return 0;
        }

        // Speicher kopieren
        memcpy(memFont, &m_buffer[0], memFontSize);

        pFontNode = (FONTNODE*)memFont;
        // Kontrollstrukturen initialisieren
        for ( i = 0; i < pFontNode->cnt; ++i )
        {
            pFont[i] = (FONT *)&memFont[pFontNode->ptrFontArray + pFontNode->cnt * sizeof( unsigned int ) + i * sizeof( FONT )];
        }

        // Zeichensaetze initialisieren
        Default( pFontNode->cnt + 1 );

        // Strukturen der Bestehende Zeichensaetze uebernehmen
        for ( i = 0; i < pFontNode->cnt; ++i )
        {
            *m_pFont[i] = *pFont[i];
        }
        // Neuen Textblock anfuegen
        *m_pFont[m_pFontNode->cnt - 1] = *(FONT *)buf;

        // Erstes Freies Datenbyte
        m_pFont[0]->ptrTable = m_pFontNode->ptrFontArray + m_pFontNode->cnt * sizeof( unsigned int ) + m_pFontNode->cnt * sizeof( FONT );
        m_pFont[0]->ptrCharacter = m_pFont[0]->ptrTable + 0x100;;

        // Daten der bestehenden Zeichensaetze uebernehmen
        for ( i = 0; i < pFontNode->cnt; ++i )
        {
            memcpy( &m_buffer[m_pFont[i]->ptrTable], &memFont[pFont[i]->ptrTable], 0x100 );
            memcpy( &m_buffer[m_pFont[i]->ptrCharacter], &memFont[pFont[i]->ptrCharacter], pFont[i]->size );

            // Naechsten freien Speicher zuweisen
            m_pFont[i + 1]->ptrTable = m_pFont[i]->ptrCharacter + m_pFont[i]->size;
            m_pFont[i + 1]->ptrCharacter = m_pFont[i + 1]->ptrTable + 0x100;;
        }

        // Neuen Block anfuegen
        memcpy( &m_buffer[m_pFont[i]->ptrTable], &buf[pText->ptrIDTable], pText->cnt*sizeof(unsigned short));
        m_pFont[i]->ptrCharacter = m_pFont[i]->ptrTable + pText->cnt*sizeof(unsigned short);

        memcpy( &m_buffer[m_pFont[i]->ptrCharacter], &buf[pText->ptrTextdata], pText->size);

        // Speicher wieder freigeben
        delete[] memFont;
    }
    else
    {
        //###################################
        // Als einzelnen Zeichensatz anfuegen

        // Zeichensaetze initialisieren
        Default( 1 );

        // Neuen Textblock anfuegen
        *m_pFont[0] = *(FONT*)buf;

        // Erstes Freies Datenbyte
        m_pFont[0]->ptrTable = m_pFontNode->ptrFontArray + m_pFontNode->cnt*sizeof(unsigned int) + m_pFontNode->cnt*sizeof(FONT);
        m_pFont[0]->ptrCharacter = m_pFont[0]->ptrTable + pText->cnt*sizeof(unsigned short);

        // Neuen Block anfuegen
        memcpy(&m_buffer[m_pFont[0]->ptrTable], &buf[pText->ptrIDTable], pText->cnt*sizeof(unsigned short));
        memcpy(&m_buffer[m_pFont[0]->ptrCharacter], &buf[pText->ptrTextdata], pText->size);
    }

    // BUGFIX: Anpassung der Groesse, damit die CRC-Berechnung von der Firmware
    // durchgefuehrt werden kann. Die Firmware geht von einer festen Tabellengroesse von 256 Byte aus.
    // Bei den Textdaten kann die Groesse dieser Tabelle abhaengig von der Maximal-ID variieren.
    m_pFont[i]->size += (((TEXTDATA*)m_pFont[i])->cnt * sizeof(unsigned short)) - 256;

    return 1;
}

bool FileDisplayFonts::loadBin( std::ifstream &fs, std::size_t size ) 
{
    // Lesen aus dem Stream
    bool res = true;

    if ( size == 0 )
    {
        setErrorState( Error_state::no_data );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        res = false;
    }
    else if ( fs.is_open() == false )
    {
        // Stream ist nicht geoffnet.
        setErrorState( Error_state::file_open );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        res = false;
    }
    else
    {
        try
        {
            m_buffer.resize( size );
            fs.read( reinterpret_cast<char *>( &m_buffer[0] ), size );

            int i;
            // 1.
            // Zeichensaetze pruefen

            // Zeigerarray und Zeichensatzstrukturzeiger initialisieren
            if ( m_pFontNode->cnt > max_fonts )
            {
                setErrorState( Error_state::format );
                setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                res = false;
            }
            else
            {
                for ( i = 0; i < m_pFontNode->cnt; i++ )
                    m_pFont[i] = (FONT*)&m_buffer[m_pFontNode->ptrFontArray + m_pFontNode->cnt * sizeof(unsigned int) + i * sizeof(FONT)];

                // CRC32 pruefen
                for ( i = 0; i < m_pFontNode->cnt; i++ )
                {
                    if ( m_pFont[i]->crc != crc32( &m_buffer[m_pFont[i]->ptrTable], 0x100 + m_pFont[i]->size ) )
                    {
                        setErrorState( Error_state::crc_fail );
                        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                        res = false;
                    }
                }
            }
        }
        catch ( ... )
        {
            // Fehler zuweisen
            setErrorState( Error_state::file_read );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            res = false;
        }
    }

    return res;
}

bool FileDisplayFonts::saveBin( std::ofstream &fs ) 
{
    bool res = true;

    if ( fs.is_open() == false )
    {
        // Stream nicht geoffnet.
        setErrorState( Error_state::file_open );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        res = false;
    }
    else
    {
        try
        {
            fs.write( reinterpret_cast<char *>( &m_buffer[0] ), m_buffer.size() );
        }
        catch ( ... )
        {
            // Fehler zuweisen
            setErrorState( Error_state::file_write );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            res = false;
        }
    }

    return res;
}
